# vim:et sts=4 sw=4
#
# ibus-typing-booster - A completion input method for IBus
#
# Copyright (c) 2025 Mike FABIAN <mfabian@redhat.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
'''
dictionary_download.py

Uses Gio.File to download dictionary files with a Gtk3 progress dialog.
'''
from typing import Dict
from typing import Set
from typing import List
from typing import Optional
from typing import Callable
# pylint: disable=wrong-import-position
import sys
if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal
# pylint: enable=wrong-import-position
InstallStatus = Literal['success', 'cancelled', 'failure']
OutputCallback = Callable[[str], None]
CompleteCallback = Callable[[InstallStatus], None]
import pathlib
import os
import locale
import logging
from gi import require_version
# pylint: disable=wrong-import-position
require_version('GLib', '2.0')
require_version('Gio', '2.0')
from gi.repository import GLib # type: ignore
from gi.repository import Gio # type: ignore
try:
    require_version('Soup', '3.0')
    from gi.repository import Soup # type: ignore
except Exception: # pylint: disable=broad-exception-caught
    # no Soup 3 available, try Soup 2.4
    require_version('Soup', '2.4')
    from gi.repository import Soup # type: ignore
# pylint: enable=wrong-import-position
from i18n import _, init as i18n_init
sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..', 'engine'))
# pylint: disable=import-error, wrong-import-order
import itb_util_core
from itb_gtk import Gtk
from g_compat_helpers import (
    add_child,
    show_all,
)
# pylint: enable=import-error, wrong-import-order

LOGGER = logging.getLogger('ibus-typing-booster')

# pylint: disable=line-too-long
# https://github.com/LibreOffice/dictionaries/
DICTIONARY_SOURCES = {
    'af': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/af_ZA/af_ZA',
    'af_ZA': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/af_ZA/af_ZA',
    'an': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/an_ES/an_ES',
    'an_ES': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/an_ES/an_ES',
    'ar': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/ar/ar',
    'ar_AE': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/ar/ar',
    'ar_BH': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/ar/ar',
    'ar_DJ': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/ar/ar',
    'ar_DZ': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/ar/ar',
    'ar_EG': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/ar/ar',
    'ar_IL': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/ar/ar',
    'ar_IN': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/ar/ar',
    'ar_IQ': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/ar/ar',
    'ar_JO': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/ar/ar',
    'ar_KM': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/ar/ar',
    'ar_KW': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/ar/ar',
    'ar_LB': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/ar/ar',
    'ar_LY': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/ar/ar',
    'ar_MA': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/ar/ar',
    'ar_MR': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/ar/ar',
    'ar_OM': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/ar/ar',
    'ar_PS': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/ar/ar',
    'ar_QA': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/ar/ar',
    'ar_SA': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/ar/ar',
    'ar_SD': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/ar/ar',
    'ar_SO': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/ar/ar',
    'ar_SY': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/ar/ar',
    'ar_TD': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/ar/ar',
    'ar_TN': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/ar/ar',
    'ar_YE': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/ar/ar',
    'as': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/as_IN/as_IN',
    'as_IN': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/as_IN/as_IN',
    # Fedora package: hunspell-ast-2.0-17.fc43.noarch
    # ast: (This is ZIP archive, not sure whether I should support that.
    # https://extensions.libreoffice.org/assets/downloads/340/1717185915/ort-ast-20240531.oxt
    'be': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/be_BY/be-official',
    'be_BY': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/be_BY/be-official',
    'bg': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/bg_BG/bg_BG',
    'bg_BG': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/bg_BG/bg_BG',
    'bn': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/bn_BD/bn_BD',
    'bn_BD': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/bn_BD/bn_BD',
    'bn_IN': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/bn_BD/bn_BD',
    'bo': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/bo/bo',
    'bo_CN': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/bo/bo',
    'bo_IN': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/bo/bo',
    'br': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/br_FR/br_FR',
    'br_FR': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/br_FR/br_FR',
    'bs': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/bs_BA/bs_BA',
    'bs_BA': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/bs_BA/bs_BA',
    'ca': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/ca/dictionaries/ca',
    'ca_AD': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/ca/dictionaries/ca',
    'ca_ES': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/ca/dictionaries/ca',
    'ca_ES_VALENCIA': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/ca/dictionaries/ca-valencia',
    'ca_FR': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/ca/dictionaries/ca',
    'ca_IT': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/ca/dictionaries/ca',
    'ckb': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/ckb/dictionaries/ckb',
    'ckb_IQ': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/ckb/dictionaries/ckb',
    'cs': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/cs_CZ/cs_CZ',
    'cs_CZ': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/cs_CZ/cs_CZ',
    'da': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/da_DK/da_DK',
    'da_DK': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/da_DK/da_DK',
    'de': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/de/de_DE_frami',
    'de_AT': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/de/de_AT_frami',
    'de_BE': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/de/de_DE_frami',
    'de_CH': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/de/de_CH_frami',
    'de_DE': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/de/de_DE_frami',
    'de_IT': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/de/de_AT_frami',
    'de_LI': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/de/de_CH_frami',
    'de_LU': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/de/de_DE_frami',
    'el': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/el_GR/el_GR',
    'el_CY': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/el_GR/el_GR',
    'el_GR': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/el_GR/el_GR',
    'en': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_US',
    'en_001': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_US',
    'en_150': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_AG': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_AI': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_AS': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_AT': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_AU': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_AU',
    'en_BB': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_BE': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_BI': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_BM': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_BS': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_BW': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_BZ': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_CA': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_CA',
    'en_CC': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_CH': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_CK': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_CM': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_CX': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_CY': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_DE': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_DG': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_DK': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_DM': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_GB': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_ER': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_FI': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_FJ': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_FK': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_FM': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_GD': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_GG': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_GH': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_GI': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_GM': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_GU': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_GY': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_HK': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_IE': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_IL': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_IM': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_IN': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_IO': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_JE': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_JM': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_KE': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_KI': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_KN': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_KY': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_LC': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_LR': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_LS': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_MG': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_MH': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_MO': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_MP': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_MS': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_MT': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_MU': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_MW': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_MY': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_NA': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_NF': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_NG': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_NL': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_NR': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_NU': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_NZ': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_PG': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_PH': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_PK': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_PN': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_PR': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_PW': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_RW': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_SB': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_SC': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_SD': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_SE': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_SG': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_SH': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_SI': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_SL': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_SS': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_SX': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_SZ': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_TC': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_TK': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_TO': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_TT': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_TV': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_TZ': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_UG': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_UM': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_US',
    'en_US': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_US',
    'en_US_POSIX': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_US',
    'en_VC': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_VG': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_VI': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_US',
    'en_VU': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_WS': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_ZA': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_ZA',
    'en_ZM': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'en_ZW': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/en/en_GB',
    'eo': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/eo/eo',
    'es': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/es/es_ES',
    'es_419': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/es/es_MX',
    'es_AR': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/es/es_AR',
    'es_BO': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/es/es_BO',
    'es_BR': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/es/es_UY',
    'es_BZ': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/es/es_MX',
    'es_CL': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/es/es_CL',
    'es_CO': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/es/es_CO',
    'es_CR': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/es/es_CR',
    'es_CU': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/es/es_CU',
    'es_DO': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/es/es_DO',
    'es_EA': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/es/es_ES',
    'es_EC': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/es/es_EC',
    'es_ES': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/es/es_ES',
    'es_GQ': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/es/es_GQ',
    'es_GT': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/es/es_GT',
    'es_HN': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/es/es_HN',
    'es_MX': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/es/es_MX',
    'es_NI': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/es/es_NI',
    'es_PA': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/es/es_PA',
    'es_PE': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/es/es_PE',
    'es_PH': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/es/es_PH',
    'es_PR': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/es/es_PR',
    'es_PY': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/es/es_PY',
    'es_SV': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/es/es_SV',
    'es_US': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/es/es_US',
    'es_UY': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/es/es_UY',
    'es_VE': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/es/es_VE',
    'et': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/et_EE/et_EE',
    'et_EE': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/et_EE/et_EE',
    'fa': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/fa_IR/fa-IR',
    'fa_IR': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/fa_IR/fa-IR',
    'fr': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/fr_FR/fr',
    'fr_BE': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/fr_FR/fr',
    'fr_BF': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/fr_FR/fr',
    'fr_BI': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/fr_FR/fr',
    'fr_BJ': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/fr_FR/fr',
    'fr_BL': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/fr_FR/fr',
    'fr_CA': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/fr_FR/fr',
    'fr_CD': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/fr_FR/fr',
    'fr_CF': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/fr_FR/fr',
    'fr_CG': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/fr_FR/fr',
    'fr_CH': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/fr_FR/fr',
    'fr_CI': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/fr_FR/fr',
    'fr_CM': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/fr_FR/fr',
    'fr_DJ': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/fr_FR/fr',
    'fr_DZ': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/fr_FR/fr',
    'fr_FR': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/fr_FR/fr',
    'fr_GA': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/fr_FR/fr',
    'fr_GF': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/fr_FR/fr',
    'fr_GN': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/fr_FR/fr',
    'fr_GP': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/fr_FR/fr',
    'fr_GQ': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/fr_FR/fr',
    'fr_HT': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/fr_FR/fr',
    'fr_KM': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/fr_FR/fr',
    'fr_LU': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/fr_FR/fr',
    'fr_MA': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/fr_FR/fr',
    'fr_MC': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/fr_FR/fr',
    'fr_MF': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/fr_FR/fr',
    'fr_MG': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/fr_FR/fr',
    'fr_ML': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/fr_FR/fr',
    'fr_MQ': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/fr_FR/fr',
    'fr_MR': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/fr_FR/fr',
    'fr_MU': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/fr_FR/fr',
    'fr_NC': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/fr_FR/fr',
    'fr_NE': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/fr_FR/fr',
    'fr_PF': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/fr_FR/fr',
    'fr_PM': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/fr_FR/fr',
    'fr_RE': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/fr_FR/fr',
    'fr_RW': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/fr_FR/fr',
    'fr_SC': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/fr_FR/fr',
    'fr_SN': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/fr_FR/fr',
    'fr_SY': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/fr_FR/fr',
    'fr_TD': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/fr_FR/fr',
    'fr_TG': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/fr_FR/fr',
    'fr_TN': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/fr_FR/fr',
    'fr_VU': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/fr_FR/fr',
    'fr_WF': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/fr_FR/fr',
    'fr_YT': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/fr_FR/fr',
    'gd': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/gd_GB/gd_GB',
    'gd_GB': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/gd_GB/gd_GB',
    'gl': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/gl/gl_ES',
    'gl_ES': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/gl/gl_ES',
    'gu': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/gu_IN/gu_IN',
    'gu_IN': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/gu_IN/gu_IN',
    'gug': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/gug/gug',
    'he': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/he_IL/he_IL',
    'he_IL': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/he_IL/he_IL',
    'hi': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/hi_IN/hi_IN',
    'hi_IN': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/hi_IN/hi_IN',
    'hr': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/hr_HR/hr_HR',
    'hr_HR': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/hr_HR/hr_HR',
    'hu': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/hu_HU/hu_HU',
    'hu_HU': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/hu_HU/hu_HU',
    'id': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/id/id_ID',
    'id_ID': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/id/id_ID',
    'is': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/is/is',
    'is_IS': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/is/is',
    'it': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/it_IT/it_IT',
    'it_CH': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/it_IT/it_IT',
    'it_IT': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/it_IT/it_IT',
    'ku_SY': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/kmr_Latn/kmr_Latn',
    'ku_TR': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/kmr_Latn/kmr_Latn',
    'kmr': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/kmr_Latn/kmr_Latn',
    'kmr_SY': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/kmr_Latn/kmr_Latn',
    'kmr_TR': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/kmr_Latn/kmr_Latn',
    'kmr_Latn': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/kmr_Latn/kmr_Latn',
    'kmr_Latn_TR': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/kmr_Latn/kmr_Latn',
    'kmr_Latn_SY': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/kmr_Latn/kmr_Latn',
    'kn': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/kn_IN/kn_IN',
    'kn_IN': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/kn_IN/kn_IN',
    'ko': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/ko_KR/ko_KR',
    'ko_KR': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/ko_KR/ko_KR',
    'lo': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/lo_LA/lo_LA',
    'lo_LA': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/lo_LA/lo_LA',
    'lt': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/lt_LT/lt',
    'lt_LT': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/lt_LT/lt',
    'lv': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/lv_LV/lv_LV',
    'lv_LV': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/lv_LV/lv_LV',
    # 'mg_MG' Fedora package: hunspell-mg-0.20050109-36.fc43.noarch
    # http://download.services.openoffice.org/contrib/dictionaries/mg_MG.zip
    'mn': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/mn_MN/mn_MN',
    'mn_MN': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/mn_MN/mn_MN',
    'mr': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/mr_IN/mr_IN',
    'mr_IN': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/mr_IN/mr_IN',
    'ne': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/ne_NP/ne_NP',
    'ne_IN': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/ne_NP/ne_NP',
    'ne_NP': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/ne_NP/ne_NP',
    'nl': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/nl_NL/nl_NL',
    'nl_AW': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/nl_NL/nl_NL',
    'nl_BE': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/nl_NL/nl_NL',
    'nl_NL': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/nl_NL/nl_NL',
    'nb': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/no/nb_NO',
    'nb_NO': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/no/nb_NO',
    'nb_SJ': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/no/nb_NO',
    'nn': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/no/nn_NO',
    'nn_NO': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/no/nn_NO',
    'oc': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/oc_FR/oc_FR',
    'oc_FR': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/oc_FR/oc_FR',
    'or': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/or_IN/or_IN',
    'or_IN': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/or_IN/or_IN',
    'pa': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/pa_IN/pa_IN',
    'pa_IN': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/pa_IN/pa_IN',
    'pl': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/pl_PL/pl_PL',
    'pl_PL': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/pl_PL/pl_PL',
    'pt': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/pt_BR/pt_BR',
    'pt_BR': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/pt_BR/pt_BR',
    'pt_PT': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/pt_PT/pt_PT',
    'ro': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/ro/ro_RO',
    'ro_RO': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/ro/ro_RO',
    'ru': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/ru_RU/ru_RU',
    'ru_RU': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/ru_RU/ru_RU',
    'ru_UA': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/ru_RU/ru_RU',
    'sa': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/sa_IN/sa_IN',
    'sa_IN': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/sa_IN/sa_IN',
    'si': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/si_LK/si_LK',
    'si_LK': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/si_LK/si_LK',
    'sk': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/sk_SK/sk_SK',
    'sk_SK': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/sk_SK/sk_SK',
    'sl': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/sl_SI/sl_SI',
    'sl_SI': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/sl_SI/sl_SI',
    'sq': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/sq_AL/sq_AL',
    'sq_AL': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/sq_AL/sq_AL',
    'sr_Latn': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/sr/sr-Latn',
    'sr': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/sr/sr',
    'sr_Cyrl': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/sr/sr',
    'sr_ME': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/sr/sr',
    'sr_RS': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/sr/sr',
    'sr_YU': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/sr/sr',
    'sv': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/sv_SE/sv_SE',
    'sv_FI': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/sv_SE/sv_FI',
    'sv_SE': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/sv_SE/sv_SE',
    'sw': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/sw_TZ/sw_TZ',
    'sw_KE': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/sw_TZ/sw_TZ',
    'sw_TZ': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/sw_TZ/sw_TZ',
    'ta': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/ta_IN/ta_IN',
    'ta_IN': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/ta_IN/ta_IN',
    'te': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/te_IN/te_IN',
    'te_IN': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/te_IN/te_IN',
    'th': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/th_TH/th_TH',
    'th_TH': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/th_TH/th_TH',
    'tr': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/tr_TR/tr_TR',
    'tr_TR': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/tr_TR/tr_TR',
    'uk': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/uk_UA/uk_UA',
    'uk_UA': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/uk_UA/uk_UA',
    'vi': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/vi/vi_VN',
    'vi_VN': 'https://raw.githubusercontent.com/LibreOffice/dictionaries/master/vi/vi_VN',
}
# pylint: enable=line-too-long



def download_file_async(
    url: str,
    dest_path: str,
    on_output: Optional[Callable[[str], None]],
    on_progress: Optional[Callable[[float], None]],
    on_complete: Optional[Callable[[InstallStatus], None]],
    cancellable: Optional[Gio.Cancellable] = None,
) -> None:
    '''Download a file from url to dest_path asynchronously.

    Using libsoup async instead of Gio.File.copy_async, which is
    broken on Ubuntu for HTTP URIs.

    on_output(line):       textual log
    on_progress(fraction): 0.0..1.0 progress fraction
                           for the current file(may be coarse)
    on_complete(status):   'success' | 'failure' | 'cancelled'
    '''
    try:
        os.makedirs(os.path.dirname(dest_path), exist_ok=True)
    except Exception as error: # pylint: disable=broad-except
        LOGGER.exception('Failed to make directory: %s', error)
        if on_output:
            on_output(f'Failed to make directory: {error}')
        if on_complete and not (cancellable and cancellable.is_cancelled()):
            on_complete('failure')
    try:
        dest_file = Gio.File.new_for_path(dest_path)
        out_stream = dest_file.replace(
            None,
            False,
            Gio.FileCreateFlags.NONE, # pylint: disable=no-member
            cancellable)
    except GLib.Error as err:
        LOGGER.exception('Failed to open destination: %s', err)
        if on_output:
            on_output(f'Failed to open destination: {err}')
        if on_complete:
            on_complete('failure')
        return

    if Soup.MAJOR_VERSION == 3:
        session: 'Soup.Session' = Soup.Session()
        message: 'Soup.Message' = Soup.Message.new('GET', url)

        def on_read_finished(
                session: 'Soup.Session',
                result: 'Gio.AsyncResult',
                msg: 'Soup.Message',
        ) -> None:
            try:
                # This returns a GLib.Bytes (GBytes) or similar; in any case it's
                # the whole response body already read by libsoup.
                body = session.send_and_read_finish(result)
                data_bytes: Optional[bytes] = None
                try:
                    # GLib.Bytes.get_data()
                    data_bytes = body.get_data()
                except Exception: # pylint: disable=broad-exception-caught
                    try:
                        # Maybe it’s already a Python bytes object
                        data_bytes = bytes(body)
                    except Exception: # pylint: disable=broad-exception-caught
                        if on_output:
                            on_output('Failed: could not obtain response bytes')
                        if on_complete:
                            on_complete('failure')
                        return
                gbytes = GLib.Bytes.new(data_bytes)
                try:
                    out_stream.write_bytes(gbytes, cancellable)
                except GLib.Error as err:
                    LOGGER.exception('Write failed: %s', err)
                    if on_output:
                        on_output(f'Write failed: {err}')
                    if on_complete:
                        on_complete('failure')
                    return
                try:
                    out_stream.close(cancellable)
                except GLib.Error as err:
                    LOGGER.exception('Failed to close output: %s', err)
                    if on_output:
                        on_output(f'Failed to close output: {err}')
                    if on_complete:
                        on_complete('failure')
                    return
                if on_progress:
                    on_progress(1.0) # We’ve read the whole body
                if on_output:
                    on_output(f'{len(data_bytes)/1024.0:.1f} KiB')
                status = msg.get_status()
                if status == Soup.Status.OK:
                    if on_output:
                        on_output('Finished.')
                    if on_complete:
                        on_complete('success')
                else:
                    if on_output:
                        on_output(f'Failed: HTTP {status}')
                    if on_complete:
                        on_complete('failure')
            except GLib.Error as err:
                if cancellable and cancellable.is_cancelled():
                    if on_output:
                        on_output('Cancelled.')
                    if on_complete:
                        on_complete('cancelled')
                else:
                    LOGGER.exception('Failed to send request: %s', err)
                    if on_output:
                        on_output(f'Failed to send request: {err}')
                    if on_complete:
                        on_complete('failure')

        session.send_and_read_async(
            message,
            GLib.PRIORITY_DEFAULT,
            cancellable,
            on_read_finished,
            message)
        return

    # Soup3 could not be imported, try Soup2:
    session_soup2: 'Soup.SessionAsync' = (
        Soup.SessionAsync()) # pylint: disable=c-extension-no-member
    message_soup2: 'Soup.Message' = Soup.Message.new('GET', url)

    total_bytes: int = -1
    received_bytes: int = 0

    def on_headers(msg: 'Soup.Message') -> None:
        nonlocal total_bytes
        total_bytes = msg.response_headers.get_content_length()

    def on_chunk(_msg: 'Soup.Message', chunk: 'Soup.Buffer') -> None:
        nonlocal received_bytes
        data_bytes = chunk.get_data()
        received_bytes += len(data_bytes)
        try:
            out_stream.write(data_bytes, cancellable)
        except GLib.Error as err:
            LOGGER.exception('Write failed: %s', err)
            if on_output:
                on_output(f'Write failed: {err}')
            if on_complete:
                on_complete('failure')
            return
        if on_progress and total_bytes > 0:
            frac = min(1.0, received_bytes / total_bytes)
            on_progress(frac)
        if on_output:
            on_output(f'{received_bytes/1024.0:.1f} KiB')

    def on_finished(msg: 'Soup.Message') -> None:
        try:
            out_stream.close(cancellable)
        except GLib.Error:
            pass
        if msg.status_code == 200:
            if on_output:
                on_output('Finished.')
            if on_complete:
                on_complete('success')
            return
        if on_output:
            on_output(f'Failed: HTTP {msg.status_code}')
        if on_complete:
            on_complete('failure')

    message_soup2.connect('got_headers', on_headers)
    message_soup2.connect('got_chunk', on_chunk)
    message_soup2.connect('finished', on_finished)
    session_soup2.queue_message(message_soup2, None)

def download_dictionaries_sequentially_async(
    languages: Set[str],
    on_output: Optional[Callable[[str], None]],
    on_progress: Optional[Callable[[float], None]],
    on_complete: Optional[Callable[[InstallStatus], None]],
    cancellable: Optional[Gio.Cancellable] = None,
) -> None:
    '''Download dictionaries sequentially, one by one.'''
    url_list: List[Dict[str, str]] = []
    for language in sorted(languages):
        url_base = DICTIONARY_SOURCES.get(language, '')
        if not url_base:
            continue
        for suffix in ('.dic', '.aff'):
            url_list.append({
                'url': f'{url_base}{suffix}',
                'dest':
                itb_util_core.xdg_save_data_path('ibus-typing-booster/data/')
                + language + suffix})

    results: Dict[str, bool] = {}

    def download_next(index: int = 0) -> None:
        '''Download the next url'''
        if cancellable and cancellable.is_cancelled():
            LOGGER.info('Download cancelled, stopping sequence.')
            if on_complete:
                on_complete('cancelled')
            return
        if index >= len(url_list):
            success = all(results.values())
            LOGGER.info('All downloads done: success=%s', success)
            if on_complete:
                on_complete('success' if success else 'failure')
            return
        url = url_list[index]['url']
        destination = url_list[index]['dest']
        LOGGER.info('Downloading url %s (%d/%d)', url, index + 1, len(url_list))
        if on_output:
            on_output(f'📥 {url}...')

        def handle_output(line: str) -> None:
            '''Call on_output'''
            if on_output:
                on_output(line)

        def handle_progress(fraction: float) -> None:
            '''Call on_progress'''
            if on_progress:
                on_progress(fraction)

        def handle_complete(status: InstallStatus) -> None:
            '''Called when one url is completed

            Puts easily parsable lines into the output to make it
            easy for dialog to update the progress.
            '''
            results[url] = status == 'success'
            if on_output:
                on_output(f'{url} ➡️ {destination} {"✔️" if status == "success" else "⚠️"}')
            if status == 'cancelled':
                LOGGER.info('Installation cancelled before url #%d', index)
                return
            download_next(index + 1)

        download_file_async(
            url,
            destination,
            on_output=handle_output,
            on_progress=handle_progress,
            on_complete=handle_complete,
            cancellable=cancellable)

    download_next()

def ensure_enchant_symlinks(data_dir: str) -> None:
    '''Ensure Enchant hunspell/nuspell directories link to downloaded dictionaries.

    Creates or replaces symlinks for *.dic/*.aff in data_dir under
    ~/.config/enchant/{hunspell,nuspell}/, and cleans up broken symlinks.
    '''
    config_root = pathlib.Path.home() / '.config' / 'enchant'
    data_path = pathlib.Path(data_dir)
    if not data_path.exists():
        LOGGER.warning('Data directory does not exist: %s', data_dir)
        return
    for backend in ('hunspell', 'nuspell'):
        backend_dir = config_root / backend
        backend_dir.mkdir(parents=True, exist_ok=True)
        # Clean up broken symlinks first:
        for link in backend_dir.glob('*'):
            if link.is_symlink() and not link.exists():
                LOGGER.info('Removing broken symlink: %s', link)
                try:
                    link.unlink()
                except OSError as err:
                    LOGGER.error(
                        'Failed to remove broken symlink %s: %s', link, err)
        # Create or update valid symlinks:
        for dict_file in pathlib.Path(data_dir).glob('*.dic'):
            for suffix in ('dic', 'aff'):
                source = data_path / f'{dict_file.stem}.{suffix}'
                if not source.exists():
                    continue
                dest = backend_dir / source.name
                try:
                    # Replace only missing or symlinked destinations
                    if dest.exists():
                        if dest.is_symlink():
                            dest.unlink()
                        else:
                            LOGGER.info(
                                'Preserving existing user dictionary: %s', dest)
                            continue
                    os.symlink(source, dest)
                    LOGGER.info('Linked %s → %s', dest, source)
                except OSError as err:
                    LOGGER.error(
                        'Failed to create symlink %s → %s: %s', dest, source, err)

def download_dictionaries_with_dialog(
    parent: Optional[Gtk.Window],
    languages: Set[str],
    on_complete: Optional[CompleteCallback] = None,
) -> None:
    '''Show a transient GTK dialog to download dictionaries asynchronously.'''
    dialog = Gtk.Dialog(
        title='📥 ' + _('Download dictionaries'),
        transient_for=parent,
        modal=True,
        destroy_with_parent=True)
    dialog.set_default_size(480, 650)
    vbox = dialog.get_content_area()
    top_label = Gtk.Label(label='')
    top_label.set_xalign(0)
    add_child(vbox, top_label)
    textview = Gtk.TextView()
    textview.set_editable(False)
    textview.set_wrap_mode(Gtk.WrapMode.WORD_CHAR)
    textbuffer = textview.get_buffer()
    scrolled = Gtk.ScrolledWindow()
    scrolled.set_vexpand(True)
    scrolled.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
    add_child(scrolled, textview)
    add_child(vbox, scrolled)
    progressbar = Gtk.ProgressBar()
    progressbar.set_show_text(True)
    progressbar.set_text('0%')
    add_child(vbox, progressbar)
    action_area = Gtk.Box()
    action_area.set_orientation(Gtk.Orientation.HORIZONTAL)
    action_area.set_halign(Gtk.Align.END)
    action_area.set_valign(Gtk.Align.CENTER)
    action_area.set_can_focus(False)
    action_area.set_hexpand(True)
    action_area.set_vexpand(False)
    action_area.set_spacing(0)
    add_child(vbox, action_area)
    cancel_button = Gtk.Button.new_with_mnemonic(_('_Cancel'))
    add_child(action_area, cancel_button)
    close_button = Gtk.Button.new_with_mnemonic(_('_Close'))
    add_child(action_area, close_button)
    show_all(dialog)
    close_button.set_visible(False)

    url_index: int = 0
    url_count: int = 2 * len(languages)
    cancellable: Gio.Cancellable = Gio.Cancellable()

    def append_line(line: str) -> None:
        '''Append a line to the textbuffer showing detailed progress'''
        nonlocal url_index
        end = textbuffer.get_end_iter()
        textbuffer.insert(end, line + '\n')
        textview.scroll_mark_onscreen(textbuffer.get_insert())
        # Detect per-url completion (From messages I emit myself)
        if '✔️' in line or '⚠️' in line:
            # Count this url as fully processed
            url_index += 1
            text = top_label.get_text()
            # In Python < 3.12, f-string expression part cannot include a backslash
            newline = '\n'
            top_label.set_text(f'{text + newline if text else ""}{line}')

    def update_progressbar(fraction: float) -> None:
        '''Update the progressbar showing the total progress'''
        # Make sure UI never jumps backward if fraction temporarily
        # drops (some async backends report non-monotonic progress
        current_fraction = max(
            progressbar.get_fraction(), (url_index + fraction) / url_count)
        progressbar.set_fraction(current_fraction)
        progressbar.set_text(
            f'{url_index}/{url_count} ({int(current_fraction * 100)}%)')

    def finish(status: InstallStatus) -> None:
        '''Installation of all urls finished or cancelled.'''
        if status == 'success':
            progressbar.set_fraction(1.0)
            progressbar.set_text(f'{url_count}/{url_count} (100%)')
            ensure_enchant_symlinks(
                itb_util_core.xdg_save_data_path('ibus-typing-booster/data/'))
        elif status == 'cancelled':
            end = textbuffer.get_end_iter()
            textbuffer.insert(end, '⚠️ Download cancelled by user.\n')
        else: # 'failed'
            end = textbuffer.get_end_iter()
            textbuffer.insert(end, '❌ There were some failures.\n')
        cancel_button.set_visible(False)
        close_button.set_visible(True)
        if on_complete:
            on_complete(status)

    def on_cancel(_button: Gtk.Button) -> None:
        '''Cancel button clicked'''
        cancellable.cancel()
        cancel_button.set_visible(False)
        close_button.set_visible(True)

    def on_close(_button: Gtk.Button) -> None:
        '''Close button clicked'''
        if glib_main_loop is not None and glib_main_loop.is_running():
            glib_main_loop.quit()
        dialog.destroy()

    cancel_button.connect('clicked', on_cancel)
    close_button.connect('clicked', on_close)
    download_dictionaries_sequentially_async(
        languages,
        on_output=append_line,
        on_progress=update_progressbar,
        on_complete=finish,
        cancellable=cancellable)
    glib_main_loop: Optional[GLib.MainLoop] = None
    # If a GLib main loop is already running (e.g. from setup UI),
    # don’t start another one; just return control to it.
    if GLib.main_depth() > 0: # pylint: disable=no-value-for-parameter
        # There’s already a running loop
        return
    # Otherwise (standalone testing), run our own main loop
    glib_main_loop = GLib.MainLoop()
    glib_main_loop.run()
    dialog.destroy()

if __name__ == '__main__':
    LOG_HANDLER = logging.StreamHandler(stream=sys.stderr)
    LOGGER.setLevel(logging.DEBUG)
    LOGGER.addHandler(LOG_HANDLER)

    try:
        locale.setlocale(locale.LC_ALL, '')
    except locale.Error:
        LOGGER.exception("IBUS-WARNING **: Using the fallback 'C' locale")
        locale.setlocale(locale.LC_ALL, 'C')
    i18n_init()

    def finish_test(status: InstallStatus) -> None:
        '''Called when downloads have completed'''
        if status == 'success':
            LOGGER.info('Install completed without errors.')
        elif status == 'cancelled':
            LOGGER.info('⚠️ Installation cancelled by user.')
        else: # 'failure'
            LOGGER.info('❌ There were some failures.')

    download_dictionaries_with_dialog(
        None,
        set(sorted(DICTIONARY_SOURCES.keys())[:4]),
        on_complete=finish_test,
    )



# vim:et sts=4 sw=4
#
# ibus-typing-booster - A completion input method for IBus
#
# Copyright (c) 2012-2013 Anish Patil <apatil@redhat.com>
# Copyright (c) 2012-2016 Mike FABIAN <mfabian@redhat.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>

'''A module used by ibus-typing-booster to suggest words by using the
hunspell dictonaries.

'''
from types import ModuleType
from typing import Optional
from typing import Dict
from typing import Tuple
from typing import List
from typing import Iterable
from typing import Any
import os
import sys
import unicodedata
import functools
import logging
import itb_util_core

LOGGER = logging.getLogger('ibus-typing-booster')

DEBUG_LEVEL = int(0)

enchant: Optional[ModuleType]
try:
    import enchant as _enchant  # type: ignore[import-untyped]
    enchant = _enchant
except ImportError:
    enchant = None

hunspell: Optional[ModuleType]
try:
    import hunspell as _hunspell  # type: ignore[import-not-found]  # ty: ignore[unresolved-import]
    hunspell = _hunspell
except ImportError:
    hunspell = None

libvoikko: Optional[ModuleType]
try:
    import libvoikko as _libvoikko  # type: ignore[import-untyped]
    libvoikko = _libvoikko
except ImportError:
    libvoikko = None

# Maximum words that should be returned.
# This should a rather big number in order not
# to throw away useful matches. But making it very huge
# makes the performance worse. For example when setting
# it to 1000, I see a noticable delay when typing the first
# letter of a word until the candidate lookup table pops up.
MAX_WORDS = 100

# pylint: disable=attribute-defined-outside-init
class Dictionary():
    '''A class to hold a hunspell dictionary'''
    # As soon as it is OK ot require Python >= 3.7, use these nicer type hints:
    #
    # _instances: Dict[Tuple[Type['Dictionary'], str], 'Dictionary'] = {}
    #
    # def __new__(cls: Type['Dictionary'], name: str = 'en_US') -> 'Dictionary':
    #
    # openSUSE Leap 15.4 still has only Python 3.6.
    _instances: Dict[Tuple[Any, str], Any] = {}

    def __new__(cls: Any, name: str = 'en_US') -> Any:
        '''Caching instances of this class and reuse previously created instances'''
        key = (cls, name)
        if key not in cls._instances:
            instance = super().__new__(cls)
            cls._instances[key] = instance
            instance._initialize(name=name)
        return cls._instances[key]

    def _initialize(self, name: str = 'en_US') -> None:
        '''Initialize a Dictionary'''
        if DEBUG_LEVEL > 1:
            LOGGER.debug('name=%s', name)
        self.name = name
        self.language = self.name.split('_')[0]
        self.dic_path = ''
        self.encoding = 'UTF-8'
        self.words: List[str]= []
        self.word_pairs: List[Tuple[str, str]] = []
        self.max_word_len = 0 # maximum length of words in this dictionary
        self.enchant_dict = None
        self.pyhunspell_object = None
        self.voikko: Optional['_libvoikko.Voikko'] = None
        if self.name != 'None':
            self.load_dictionary()

    def load_dictionary(self) -> None:
        '''Load a hunspell dictionary and instantiate a
        enchant.Dict() or a hunspell.Hunspell() object.

        '''
        if DEBUG_LEVEL > 0:
            LOGGER.debug('load_dictionary() ...\n')
        (self.dic_path,
         self.encoding,
         self.words) = itb_util_core.get_hunspell_dictionary_wordlist(self.name)
        if self.words:
            if self.language in itb_util_core.ACCENT_LANGUAGES:
                self.word_pairs = [
                    (x, itb_util_core.remove_accents(
                        x, keep=itb_util_core.ACCENT_LANGUAGES[self.language]))
                    for x in self.words
                ]
            for word in self.words:
                self.max_word_len = max(self.max_word_len, len(word))
            if DEBUG_LEVEL > 1:
                LOGGER.debug(
                    'max_word_len = %s\n', self.max_word_len)
            if self.name.split('_')[0] == 'fi':
                self.enchant_dict = None
                self.pyhunspell_object = None
                self.voikko = None
                if libvoikko is None:
                    LOGGER.warning(
                        'Language is “fi” but “import libvoikko” failed.')
                    return
                try:
                    self.voikko = libvoikko.Voikko('fi')
                except (libvoikko.VoikkoException,) as error:
                    LOGGER.warning('Init of voikko failed: %s: %s',
                                   error.__class__.__name__, error)
                    self.voikko = None
                return
            if enchant is not None:
                broker = enchant.Broker()
                broker.set_ordering(self.name, 'hunspell,nuspell,aspell,voikko')
                try:
                    self.enchant_dict = broker.request_dict(self.name)
                    if DEBUG_LEVEL > 1 and self.enchant_dict is not None:
                        LOGGER.debug(
                            '%s: %r', self.name, self.enchant_dict.provider)
                except enchant.errors.DictNotFoundError as error:
                    LOGGER.exception(
                        'Error initializing enchant for %s: %s: %s',
                        self.name, error.__class__.__name__, error)
                    self.enchant_dict = None
                except Exception as error: # pylint: disable=broad-except
                    LOGGER.exception(
                        'Unexpected error initializing enchant for %s: %s: %s',
                        self.name, error.__class__.__name__, error)
                    self.enchant_dict = None
            elif hunspell is not None and self.dic_path:
                aff_path = self.dic_path.replace('.dic', '.aff')
                try:
                    self.pyhunspell_object = (
                        hunspell.HunSpell( # pylint: disable=used-before-assignment
                            self.dic_path, aff_path))
                except hunspell.HunSpellError as error:
                    LOGGER.debug(
                        'Error initializing hunspell for %s: %s: %s',
                        self.name, error.__class__.__name__, error)
                    self.pyhunspell_object = None
                except Exception as error: # pylint: disable=broad-except
                    LOGGER.debug(
                        'Unexpected error initializing hunspell for '
                        '%s: %s: %s',
                        self.name, error.__class__.__name__, error)
                    self.pyhunspell_object = None

    def spellcheck_enchant(self, word: str) -> bool:
        '''
        Spellcheck a word using enchant

        :param word: The word to spellcheck
        :return: True if spelling is correct, False if not or unknown
        '''
        if not self.enchant_dict:
            return False
        # enchant does the right thing for all languages, including
        # Korean, if the input is a Unicode string in NFC.
        return self.enchant_dict.check(unicodedata.normalize('NFC', word))

    def spellcheck_pyhunspell(self, word: str) -> bool:
        '''
        Spellcheck a word using pyhunspell

        :param word: The word to spellcheck
        :return: True if spelling is correct, False if not or unknown
        '''
        if not self.pyhunspell_object:
            return False
        # pyhunspell needs its input passed in dictionary encoding.
        # and also returns in dictionary encoding.
        return self.pyhunspell_object.spell(
            unicodedata.normalize('NFC', word).encode(
                self.encoding, 'replace'))

    def spellcheck_voikko(self, word: str) -> bool:
        '''
        Spellcheck a word using voikko

        :param word: The word to spellcheck
        :return: True if spelling is correct, False if not or unknown
        '''
        if not self.voikko:
            return False
        # voikko works correctly if the input is a Unicode string in NFC.
        return bool(self.voikko.spell(unicodedata.normalize('NFC', word)))

    def spellcheck(self, word: str) -> bool:
        '''
        Spellcheck a word using enchant, pyhunspell, or voikko

        :param word: The word to spellcheck
        :return: True if spelling is correct, False if not or unknown

        >>> d = Dictionary('en_US')
        >>> d.spellcheck('winter')
        True

        >>> d.spellcheck('winxer')
        False

        >>> d = Dictionary('None')
        >>> d.spellcheck('winter')
        False

        >>> d.spellcheck('winxer')
        False
        '''
        if self.enchant_dict:
            return self.spellcheck_enchant(word)
        if self.pyhunspell_object:
            return self.spellcheck_pyhunspell(word)
        if self.voikko:
            return bool(self.voikko.spell(word))
        return False

    def has_spellchecking(self) -> bool:
        '''
        Returns wether this dictionary supports spellchecking or not

        :return: True if this dictionary spports spellchecking, False if not

        Examples:

        >>> d = Dictionary('en_US')
        >>> d.has_spellchecking()
        True

        >>> d = Dictionary('zh_CN')
        >>> d.has_spellchecking()
        False

        >>> d = Dictionary('None')
        >>> d.has_spellchecking()
        False
        '''
        if self.enchant_dict or self.pyhunspell_object or self.voikko:
            return True
        return False

    def spellcheck_suggest_enchant(self, word: str) -> List[str]:
        '''
        Return spellchecking suggestions for word using enchant

        :param word: The word to return spellchecking suggestions for
        :return: List of spellchecking suggestions, possibly empty.
        '''
        if not word or not self.enchant_dict:
            return []
        # enchant does the right thing for all languages, including
        # Korean, if the input is NFC. It takes Unicode strings and
        # returns Unicode strings, no encoding and decoding is
        # necessary, neither for Python2 nor for Python3.
        return [
            unicodedata.normalize(
                itb_util_core.NORMALIZATION_FORM_INTERNAL, x)
            for x in
            self.enchant_dict.suggest(unicodedata.normalize('NFC', word))
            ]

    def spellcheck_suggest_pyhunspell(self, word: str) -> List[str]:
        '''
        Return spellchecking suggestions for word using pyhunspell

        :param word: The word to return spellchecking suggestions for
        :return: List of spellchecking suggestions, possibly empty.
        '''
        if not word or not self.pyhunspell_object:
            return []
        # pyhunspell needs its input passed in dictionary encoding.
        return [
            unicodedata.normalize(
                itb_util_core.NORMALIZATION_FORM_INTERNAL, x)
            for x in
            self.pyhunspell_object.suggest(
                unicodedata.normalize('NFC', word).encode(
                    self.encoding, 'replace'))
            ]

    def spellcheck_suggest_voikko(self, word: str) -> List[str]:
        '''
        Return spellchecking suggestions for word using voikko

        :param word: The word to return spellchecking suggestions for
        :return: List of spellchecking suggestions, possibly empty.
        '''
        if not word or not self.voikko:
            return []
        return [
            unicodedata.normalize(
                itb_util_core.NORMALIZATION_FORM_INTERNAL, x)
            for x in
            self.voikko.suggest(unicodedata.normalize('NFC', word))
            ]

    # Don’t use @lru_cache(maxsize=None) here, that has a high risk of
    # memory leaks.  It caches forever — and it keeps strong
    # references to all function arguments and results. If the method
    # is on a class instance (self), and the cache calls involving
    # self, then self gets kept alive — even if no other code
    # references it! That is a high risk of memory leaks when
    # instantiated class objects go out of scope.  With a bounded
    # cache, Python will evict the oldest cache entries automatically
    # when the cache grows beyond 500,000 entries. That is much safer
    # The self reference can still stay around as long as there are
    # still entries in the cache for that instance. But if all entries
    # referring to a self are evicted, then self can be garbage
    # collected properly.
    @functools.lru_cache(maxsize=500_000)
    def spellcheck_suggest(self, word: str) -> List[str]:
        '''Return spellchecking suggestions for word using enchant,
        pyhunspell or voikko

        :param word: The word to return spellchecking suggestions for
        :return: List of spellchecking suggestions, possibly empty.

        Results can be quite different depending on whether enchant or
        pyhunspell is used and in case of enchant whether hunspell,
        aspell, myspell, or ispell are used. So for the doctests,
        better don’t check the complete list of suggestions returned.

        Examples:

        >>> d = Dictionary('en_US')
        >>> 'Camel' in d.spellcheck_suggest('kamel')
        True

        >>> d.spellcheck_suggest('')
        []

        >>> d = Dictionary('None')
        >>> d.spellcheck_suggest('kamel')
        []

        '''
        if self.enchant_dict:
            return self.spellcheck_suggest_enchant(word)
        if self.pyhunspell_object:
            return self.spellcheck_suggest_pyhunspell(word)
        if self.voikko:
            return self.spellcheck_suggest_voikko(word)
        return []
# pylint: enable=attribute-defined-outside-init

class Hunspell:
    '''A class to suggest completions or corrections
    using a list of Hunspell dictionaries
    '''
    def __init__(self, dictionary_names: Iterable[str] = ()) -> None:
        global DEBUG_LEVEL # pylint: disable=global-statement
        try:
            DEBUG_LEVEL = int(
                str(os.getenv('IBUS_TYPING_BOOSTER_DEBUG_LEVEL')))
        except (TypeError, ValueError):
            DEBUG_LEVEL = int(0)
        if DEBUG_LEVEL > 1:
            if dictionary_names:
                LOGGER.debug(
                    'Hunspell.__init__(dictionary_names=%s)\n',
                    dictionary_names)
            else:
                LOGGER.debug('Hunspell.__init__(dictionary_names=())\n')
        self._suggest_cache: Dict[str, List[Tuple[str, int]]] = {}
        self._dictionary_names: List[str] = list(dictionary_names)
        self._dictionaries: List[Dictionary] = []
        self.init_dictionaries()

    def init_dictionaries(self) -> None:
        '''Initialize the hunspell dictionaries
        '''
        if DEBUG_LEVEL > 1:
            if self._dictionary_names:
                LOGGER.debug(
                    'Hunspell.init_dictionaries() dictionary_names=%s\n',
                    self._dictionary_names)
            else:
                LOGGER.debug(
                    'Hunspell.init_dictionaries() dictionary_names=()\n')
        self.suggest.cache_clear()
        self._dictionaries = []
        for dictionary_name in self._dictionary_names:
            self._dictionaries.append(Dictionary(name=dictionary_name))

    def get_dictionary_names(self) -> List[str]:
        '''Returns a copy of the list of dictionary names.

        It is important to return a copy, we do not want to change
        the private member variable directly.'''
        return list(self._dictionary_names[:])

    def set_dictionary_names(self, dictionary_names: List[str]) -> None:
        '''Sets the list of dictionary names.

        If the new list of dictionary names differs from the existing
        one, re-initialize the dictionaries.
        '''
        if dictionary_names != self._dictionary_names:
            if set(dictionary_names) != set(self._dictionary_names):
                # Some dictionaries are really different, reinitialize:
                self._dictionary_names = dictionary_names
                self.init_dictionaries()
            else:
                # Only the order of dictionaries has changed.
                # Reinitializing wastes time, just reorder the
                # dictionaries:
                self._dictionary_names = dictionary_names
                dictionaries_new = []
                for name in dictionary_names:
                    for dictionary in self._dictionaries:
                        if dictionary.name == name:
                            dictionaries_new.append(dictionary)
                self._dictionaries = dictionaries_new
        if DEBUG_LEVEL > 1:
            LOGGER.debug('set_dictionary_names(%s):\n', dictionary_names)
            for dictionary in self._dictionaries:
                LOGGER.debug('%s\n', dictionary.name)

    def spellcheck(self, input_phrase: str) -> bool:
        '''
        Checks if a string is likely to be spelled correctly checking
        multiple dictionaries

        :param input_phrase: A string to spellcheck
        :return: True if it is more likely to be spelled correctly,
                 False if it is more likely to be spelled incorrectly.
                 In detail this means:
                 True:
                     - If it is a correctly spelled word in at least one of
                       the dictionaries supporting spellchecking
                     - None of the dictionaries support spellchecking
                     - Contains spaces, spellchecking cannot work
                 else False.

        Examples:

        >>> h = Hunspell(['en_US', 'de_DE', 'ja_JP'])
        >>> h.spellcheck('Hello')
        True

        >>> h.spellcheck('Grüße')
        True

        >>> h.spellcheck('Gruße')
        False

        >>> h = Hunspell(['en_US', 'ja_JP'])
        >>> h.spellcheck('Grüße')
        False

        >>> h = Hunspell(['ja_JP'])
        >>> h.spellcheck('Grüße')
        True

        >>> h = Hunspell(['en_US', 'None'])
        >>> h.spellcheck('Grüße')
        False

        >>> h = Hunspell(['None'])
        >>> h.spellcheck('Grüße')
        True
        '''
        if ' ' in input_phrase:
            return True
        spellchecking_dictionaries_available = False
        spellcheck_total = False
        for dictionary in self._dictionaries:
            if dictionary.has_spellchecking():
                spellchecking_dictionaries_available = True
                spellcheck_total |= dictionary.spellcheck(input_phrase)
        if not spellcheck_total and spellchecking_dictionaries_available:
            return False
        return True

    def spellcheck_match_list(self, input_phrase: str) -> List[str]:
        '''
        Returns a list of dictionaries where input_phrase can be found

        :param input_phrase: The word to be spellchecked
        :return: A list of dictionary names which accept input phrase
                 as a valid word.

        Examples:

        >>> h = Hunspell(['en_US', 'None', 'it_IT', 'fr_FR'])
        >>> h.spellcheck_match_list('arrive')
        ['en_US', 'fr_FR']
        >>> h.spellcheck_match_list('arrivé')
        ['fr_FR']
        >>> h.spellcheck_match_list('ragazzo')
        ['it_IT']
        >>> h.spellcheck_match_list(' \t')
        []
        '''
        match_list = []
        if not input_phrase.strip():
            return []
        for dictionary in self._dictionaries:
            if dictionary.spellcheck(input_phrase):
                match_list.append(dictionary.name)
        return match_list

    def spellcheck_single_dictionary(self, words: Iterable[str] = ()) -> List[str]:
        '''
        Checks whether there is at least one dictionary where all words
        in the input list spellcheck as True.

        :param words: A list or tuple of words to spellcheck
        :return: - '' if at least one input word is empty
                 - name of a dictionary where all
                   words in the input list spellcheck as True, if
                   there is no such dictionary, return ''.

        Examples:

        >>> h = Hunspell(['en_US', 'fr_FR'])

        All 3 words spellcheck True in the French dictionary:

        >>> h.spellcheck_single_dictionary(('Je', 'suis', 'arrivé'))
        ['fr_FR']

        Not a correct French sentence but all 3 words spellcheck True
        in the French dictionary:

        >>> h.spellcheck_single_dictionary(('Je', 'suis', 'arrive'))
        ['fr_FR']

        All words spellcheck as True in both the English and the French dictionary:

        >>> h.spellcheck_single_dictionary(('arrive', 'arrive', 'arrive'))
        ['en_US', 'fr_FR']

        Not all 3 words spellcheck True in the French dictionary and
        not all 3 words spellcheck True in the English dictionary either:

        >>> h.spellcheck_single_dictionary(('Je', 'suis', 'arrived'))
        []

        All 3 words spellcheck True in the English dictionary:

        >>> h.spellcheck_single_dictionary(('I', 'have', 'arrived'))
        ['en_US']

        Unfortunately these 3 words spellcheck True in the French dictionary
        ('have' is a form of the verb 'haver', 'I' is the name of the letter):

        >>> h.spellcheck_single_dictionary(('I', 'have', 'arrivé'))
        ['fr_FR']

        If the input contains empty words the result is '':

        >>> h.spellcheck_single_dictionary(('I', 'have', ''))
        []
        '''
        if not words:
            return []
        for word in words:
            if not word:
                return []
        dictionary_names = []
        for dictionary in self._dictionaries:
            spellcheck_total = True
            for word in words:
                spellcheck_total &= dictionary.spellcheck(word)
            if spellcheck_total:
                dictionary_names.append(dictionary.name)
        return sorted(dictionary_names)

    # Don’t use @lru_cache(maxsize=None) here, that has a high risk of
    # memory leaks.  It caches forever — and it keeps strong
    # references to all function arguments and results. If the method
    # is on a class instance (self), and the cache calls involving
    # self, then self gets kept alive — even if no other code
    # references it! That is a high risk of memory leaks when
    # instantiated class objects go out of scope.  With a bounded
    # cache, Python will evict the oldest cache entries automatically
    # when the cache grows beyond 500,000 entries. That is much safer
    # The self reference can still stay around as long as there are
    # still entries in the cache for that instance. But if all entries
    # referring to a self are evicted, then self can be garbage
    # collected properly.
    @functools.lru_cache(maxsize=500_000)
    def suggest(self, input_phrase: str) -> List[Tuple[str, int]]:
        # pylint: disable=line-too-long
        '''Return completions or corrections for the input phrase

        :param input_phrase: A string to find completions or corrections for

        Returns a list of tuples of the form (<word>, <score>)
                <score> can have these values:
                    0: This is a completion, i.e. input_phrase matches
                       the beginning of <word> (accent insensitive match)
                   negative number: This is a spell checking correction

        Examples:

        (Attention, the return values are in internal
        normalization form ('NFD'))

        >>> h = Hunspell(['de_DE', 'cs_CZ'])
        >>> h.suggest('Geschwindigkeitsubertre')[0]
        ('Geschwindigkeitsübertretungsverfahren', 0)

        >>> h.suggest('Geschwindigkeitsübertretungsverfahren')[0]
        ('Geschwindigkeitsübertretungsverfahren', 0)

        >>> h.suggest('Glühwürmchen')[0]
        ('Glühwürmchen', 0)

        >>> h.suggest('Alpengluhen')[0]
        ('Alpenglühen', 0)

        >>> h.suggest('filosofictejs')[0]
        ('filosofičtější', 0)

        >>> h.suggest('filosofičtější')[0]
        ('filosofičtější', 0)

        >>> h.suggest('filosofičtějš')[0]
        ('filosofičtější', 0)

        >>> h = Hunspell(['it_IT'])
        >>> h.suggest('principianti')
        [('principianti', 0), ('principiati', -2), ('principiante', -3), ('principiarti', -4), ('principiasti', -5)]

        >>> h = Hunspell(['es_ES'])
        >>> h.suggest('teneis')
        [('tenéis', 0), ('teneos', -2), ('tenes', -3), ('tenis', -4), ('tienes', -5), ('te neis', -6), ('te-neis', -7)]

        >>> h.suggest('tenéis')[0]
        ('tenéis', 0)

        >>> h = Hunspell(['en_US'])
        >>> ('Camelot', 0) in h.suggest('camel') # Should work with aspell and hunspell
        True

        >>> h = Hunspell(['fr_FR'])
        >>> h.suggest('differemmen')
        [('différemment', 0)]

        >>> h = Hunspell(['None'])
        >>> h.suggest('camel')
        []

        >>> h = Hunspell(['None', 'en_US'])
        >>> ('Camelot', 0) in h.suggest('camel') # Should work with aspell and hunspell
        True

        >>> h = Hunspell(['en_US'])
        >>> all((x, 0) in h.suggest('tree') for x in ('trees', 'treed'))
        True

        For benchmarking:

        >>> result = [h.suggest(c) for c in 'abcdefghijklmnopqrstuvwxyz']
        >>> result = [h.suggest(word) for word in ('infl', 'mess', 'tree', 'temp', 'wood', 'gard', 'came', 'wind', 'wate', 'succ', 'dict', 'load', 'enco', 'spell', 'word', 'auto', 'mode', 'bicy', 'moun', 'batt', 'tabl', 'star', 'moon', 'univ')]
        '''
        # pylint: enable=line-too-long
        if DEBUG_LEVEL > 1:
            LOGGER.debug(
                "Hunspell.suggest() input_phrase=%(ip)s\n",
                {'ip': input_phrase.encode('UTF-8')})
        # http://pwet.fr/man/linux/fichiers_speciaux/hunspell says:
        #
        # > A dictionary file (*.dic) contains a list of words, one per
        # > line. The first line of the dictionaries (except personal
        # > dictionaries) contains the word count. Each word may
        # > optionally be followed by a slash ("/") and one or more
        # > flags, which represents affixes or special attributes.
        #
        # I.e. if '/' is already contained in the input, it cannot
        # match a word in the dictionary and we return an empty list
        # immediately:
        if not input_phrase or '/' in input_phrase:
            return []
        # make sure input_phrase is in the internal normalization form (NFD):
        input_phrase = unicodedata.normalize(
            itb_util_core.NORMALIZATION_FORM_INTERNAL, input_phrase)

        suggested_words: Dict[str, Dict[str, int]] = {}
        for dictionary in self._dictionaries:
            name = dictionary.name
            suggested_words[name] = {}
            if dictionary.words:
                if dictionary.word_pairs:
                    input_phrase_no_accents = itb_util_core.remove_accents(
                        input_phrase,
                        keep=itb_util_core.ACCENT_LANGUAGES[dictionary.language])
                # If the input phrase is longer than than the maximum
                # word length in a dictionary, don’t try
                # complete it, it just wastes time then.
                if len(input_phrase) <= dictionary.max_word_len:
                    if dictionary.word_pairs:
                        query = input_phrase_no_accents.casefold()
                        for word, stripped in dictionary.word_pairs:
                            if stripped.casefold().startswith(query):
                                suggested_words[name][word] = 0
                    else:
                        query = input_phrase.casefold()
                        for word in dictionary.words:
                            if word.casefold().startswith(query):
                                suggested_words[name][word] = 0
                if len(input_phrase) >= 4:
                    if dictionary.spellcheck(input_phrase):
                        # This is a valid word in this dictionary.
                        # It might have been missed by the
                        # matching above because the dictionary
                        # might not contain all possible word
                        # forms (The prefix and suffix information
                        # has been ignored). But the spell checker
                        # knows about this, if the spell checker
                        # thinks it is a correct word, it must be
                        # counted as a match of course:
                        suggested_words[name][input_phrase] = 0
                    # Add suffixes:
                    # See: https://github.com/mike-fabian/ibus-typing-booster/issues/799
                    suffixed_suggestions = {
                        x: 0
                        for suggestion in sorted(suggested_words[name])[0:4]
                        for x in dictionary.spellcheck_suggest(suggestion)
                        if x.startswith(suggestion)
                    }
                    suggested_words[name].update(suffixed_suggestions)
                    extra_suggestions = (
                        unicodedata.normalize(
                            itb_util_core.NORMALIZATION_FORM_INTERNAL, x)
                        for x in
                        dictionary.spellcheck_suggest(input_phrase)
                    )
                    for index, suggestion in enumerate(extra_suggestions):
                        if suggestion not in suggested_words[name]:
                            if (dictionary.word_pairs
                                and
                                itb_util_core.remove_accents(
                                    suggestion,
                                    keep=itb_util_core.ACCENT_LANGUAGES[
                                        dictionary.language])
                                == input_phrase_no_accents):
                                suggested_words[name][suggestion] = 0
                            else:
                                suggested_words[name][suggestion] = -(index + 1)
        suggested_words_total = itb_util_core.merge_dicts_max(
            *suggested_words.values())
        sorted_suggestions = sorted(
            suggested_words_total.items(),
            key=lambda x: (
                - x[1],    # 0: in dictionary, negative: spellcheck
                len(x[0]), # length of word ascending
                x[0],      # alphabetical
            ))[0:MAX_WORDS]
        return sorted_suggestions

BENCHMARK = True

def main() -> None:
    '''
    Used for testing and profiling.

    “python3 hunspell_suggest.py”

    runs some tests and prints profiling data.
    '''
    log_handler = logging.StreamHandler(stream=sys.stderr)
    log_formatter = logging.Formatter(
        '%(asctime)s %(filename)s '
        'line %(lineno)d %(funcName)s %(levelname)s: '
        '%(message)s')
    log_handler.setFormatter(log_formatter)
    LOGGER.setLevel(logging.DEBUG)
    LOGGER.addHandler(log_handler)

    if BENCHMARK:
        import cProfile # pylint: disable=import-outside-toplevel
        import pstats # pylint: disable=import-outside-toplevel
        profile = cProfile.Profile()
        profile.enable()

    import doctest # pylint: disable=import-outside-toplevel
    (failed, _attempted) = doctest.testmod()

    if BENCHMARK:
        profile.disable()
        stats = pstats.Stats(profile)
        stats.strip_dirs()
        stats.sort_stats('cumulative')
        stats.print_stats('hunspell', 25)
        stats.print_stats('enchant', 25)

    LOGGER.info('itb_util_core.remove_accents() cache info: %s',
                itb_util_core.remove_accents.cache_info())

    sys.exit(failed)

if __name__ == "__main__":
    main()



# vim:et sts=4 sw=4
#
# ibus-typing-booster - A completion input method for IBus
#
# Copyright (c) 2015-2025 Mike FABIAN <mfabian@redhat.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
'''
A module with utilites to use ollama
'''

from typing import Optional
from typing import List
from typing import Dict
from typing import Union
from typing import Any
from types import FrameType
import sys
import os
import signal
import argparse
import locale
import threading
import gettext
import logging

from gi import require_version
# pylint: disable=wrong-import-position
require_version('GLib', '2.0')
from gi.repository import GLib # type: ignore
# pylint: enable=wrong-import-position

# set_prgname before importing other modules to show the name in warning
# messages when import modules are failed. E.g. Gtk.
GLib.set_application_name('Ollama Pull')

import itb_ollama
from itb_gtk import Gtk, GTK_MAJOR
from g_compat_helpers import (
    add_child,
    set_border_width,
    show_all,
    CompatButton,
)

LOGGER = logging.getLogger('ibus-typing-booster')

glib_main_loop: Optional[GLib.MainLoop] = None

DOMAINNAME = 'ibus-typing-booster'

def _(text: str) -> str:
    '''Gettext translation function.'''
    return gettext.dgettext(DOMAINNAME, text)

def N_(text: str) -> str: # pylint: disable=invalid-name
    '''Mark string for translation without actually translating.

    Used by gettext tools to extract strings that need translation.
    '''
    return text

def str2bool(v: Union[bool, str]) -> bool:
    '''
    Convert a string to a boolean for argparse.
    '''
    if isinstance(v, bool):
        return v
    if v.lower() in ('yes', 'true', 't', '1'):
        return True
    if v.lower() in ('no', 'false', 'f', '0'):
        return False
    raise argparse.ArgumentTypeError('Boolean value expected.')

def parse_args() -> Any:
    '''
    Parse the command line arguments.
    '''
    parser = argparse.ArgumentParser(
        description='UI to show progress while pulling a ollama model.')
    parser.add_argument(
        '-m', '--model',
        nargs='?',
        type=str,
        action='store',
        default=None,
        help=('Choose the model to pull. '
              'default: "%(default)s"'))
    parser.add_argument(
        '-d', '--debug',
        nargs='?',
        type=str2bool,
        action='store',
        const=True, # if -d or --debug with no value assume True
        default=False,
        help=('Print some debug output to stdout. '
              'default: %(default)s'))
    return parser.parse_args()

_ARGS = parse_args()

class OllamaPullUI(Gtk.Window):  # type: ignore[misc]
    '''
    UI to show progress while pulling a ollama model.
    '''
    def __init__(self, model: str = '') -> None:
        self._model = model
        if self._model == '':
            raise ValueError('Model name should not be empty.')
        self._thread: Optional[threading.Thread] = None
        self._stop_event: threading.Event = threading.Event()
        self._status = ''
        self._status_labels: List[Gtk.Label] = []
        self._status_progress_bars: List[Gtk.ProgressBar] = []
        title = f'📥 {self._model}'
        Gtk.Window.__init__(self, title=title)
        self.set_title(title)
        self.set_name('Ollama Pull')
        self.set_modal(True)
        if GTK_MAJOR >= 4:
            self.connect('close-request', self._on_close)
        else:
            self.connect('delete-event', self._on_close)
        main_container = Gtk.Box()
        main_container.set_orientation(Gtk.Orientation.VERTICAL)
        main_container.set_spacing(0)
        main_container.set_hexpand(True)
        main_container.set_vexpand(True)
        add_child(self, main_container)
        self._progress_grid = Gtk.Grid()
        self._progress_grid.set_visible(True)
        self._progress_grid.set_can_focus(False)
        set_border_width(self._progress_grid, 5)
        self._progress_grid.set_row_spacing(5)
        self._progress_grid.set_column_spacing(10)
        self._progress_grid.set_row_homogeneous(False)
        self._progress_grid.set_column_homogeneous(False)
        self._progress_grid.set_hexpand(True)
        self._progress_grid.set_vexpand(False)
        self._row = -1
        add_child(main_container, self._progress_grid)
        dialog_action_area = Gtk.Box()
        dialog_action_area.set_orientation(Gtk.Orientation.HORIZONTAL)
        dialog_action_area.set_visible(True)
        dialog_action_area.set_can_focus(False)
        dialog_action_area.set_hexpand(True)
        dialog_action_area.set_vexpand(False)
        dialog_action_area.set_spacing(0)
        add_child(main_container, dialog_action_area)
        empty_hexpanding_label = Gtk.Label()
        empty_hexpanding_label.set_hexpand(True)
        empty_hexpanding_label.set_vexpand(False)
        add_child(dialog_action_area, empty_hexpanding_label)
        self._cancel_button = CompatButton(
            label=_('_Cancel'), use_underline=True)
        self._cancel_button.connect('clicked', self._on_cancel_clicked)
        add_child(dialog_action_area, self._cancel_button)
        self._close_button = CompatButton(
            label=_('_Close'), use_underline=True)
        self._close_button.connect('clicked', self._on_close_clicked)
        add_child(dialog_action_area, self._close_button)
        show_all(self)
        self._close_button.set_visible(False)
        GLib.timeout_add(200, self._check_for_interrupt)
        GLib.idle_add(self._pull)

    def _pull(self) -> None:
        '''Pull the requested model'''
        ollama_client = itb_ollama.ItbOllamaClient()
        if self._stop_event:
            self._stop_event.clear()
        self._thread = threading.Thread(
            daemon=True,
            target=ollama_client.pull,
            args=(self._model, self._pull_progress_idle_add, self._stop_event))
        self._thread.start()

    def _pull_progress_idle_add(self, progress: Dict[str, Any]) -> None:
        '''Schedule an update of the progress of pulling the ollama model'''
        GLib.idle_add(lambda:
                      self._pull_progress(progress))

    def _pull_progress(self, progress: Dict[str, Any]) -> None:
        '''Show the progress of pulling the ollama model'''
        error = progress.get('error', None)
        total = progress.get('total', None)
        completed = progress.get('completed', None)
        status = progress.get('status', None)
        if error is not None:
            LOGGER.error('Error pulling %r: %r', self._model, error)
            error_label = Gtk.Label()
            error_label.set_text(error)
            error_label.set_xalign(0)
            self._progress_grid.attach(
                error_label, 1, self._row, 1, 1)
            error_label.set_visible(True)
            self._cancel_button.set_visible(False)
            self._close_button.set_visible(True)
            return
        if status is not None and status != self._status:
            self._status = status
            self._row += 1
            self._status_labels.append(Gtk.Label())
            self._status_labels[self._row].set_xalign(0)
            self._progress_grid.attach(
                self._status_labels[self._row],
                0, self._row, 1, 1)
            self._status_labels[self._row].set_visible(True)
            self._status_progress_bars.append(Gtk.ProgressBar())
            self._status_progress_bars[self._row].set_show_text(False)
            self._status_progress_bars[self._row].set_pulse_step(0)
            # Width 150px, natural height
            self._status_progress_bars[self._row].set_size_request(150, -1)
            self._status_progress_bars[self._row].set_vexpand(True)
            self._status_progress_bars[self._row].set_valign(Gtk.Align.CENTER)
            self._status_progress_bars[self._row].set_margin_start(10)
            self._status_progress_bars[self._row].set_margin_end(10)
            self._progress_grid.attach(
                self._status_progress_bars[self._row],
                1, self._row, 1, 1)
            self._status_progress_bars[self._row].set_visible(False)
        if total is not None and completed is not None:
            fraction = float(completed / total) if total > 0.0 else 0.0
            status_text = status if status is not None else ''
            total_text = f'{total} B'
            if total > 1024**3:
                total_text = f'{total / 1024**3:.1f} GB'
            elif total > 1024**2:
                total_text = f'{total / 1024**2:.1f} MB'
            elif total > 1024:
                total_text = f'{total / 1024:.1f} kB'
            LOGGER.info('%r: %r (%.1f%%)',
                status_text, total_text, 100 * fraction)
            if self._status_labels[self._row] is not None:
                self._status_labels[self._row].set_text(
                    f'{status_text}: {total_text} ({100 * fraction:.1f}%)')
            if self._status_progress_bars[self._row] is not None:
                self._status_progress_bars[self._row].set_visible(True)
                self._status_progress_bars[self._row].set_fraction(fraction)
        elif status is not None:
            LOGGER.info('%r', status)
            if self._status_labels[self._row] is not None:
                self._status_labels[self._row].set_text(status)
        if status == 'success':
            self._cancel_button.set_visible(False)
            self._close_button.set_visible(True)

    @staticmethod
    def _quit() -> None:
        '''Quit the GLib main loop'''
        LOGGER.info('Quit GLib main loop')
        if glib_main_loop is not None:
            glib_main_loop.quit()
        else:
            raise RuntimeError('glib_main_loop not initialized!')

    def _on_close(self, *_args: Any) -> None:
        '''The window has been closed, probably by the window manager.'''
        LOGGER.info('Window deleted by the window manager.')
        self._cancel_pull()
        self.__class__._quit() # pylint: disable=protected-access

    def _cancel_pull(self) -> None:
        '''Cancel a running pull'''
        if not self._thread:
            LOGGER.info('No thread, cannot cancel.')
            return
        if not self._stop_event:
            LOGGER.info('No stop event, cannot cancel.')
            return
        if self._thread.is_alive():
            self._stop_event.set()
            self._thread.join()
            self._pull_progress({'status': 'cancelled'})
        self._stop_event.clear()
        self._thread = None
        self._cancel_button.set_visible(False)
        self._close_button.set_visible(True)

    def _on_cancel_clicked(self, *_args: Any) -> None:
        '''The button to cancel has been clicked.'''
        LOGGER.info('Cancel button clicked.')
        self._cancel_pull()

    def _on_close_clicked(self, *_args: Any) -> None:
        '''The close button has been clicked.'''
        LOGGER.info('Close clicked.')
        self._cancel_pull()
        self.__class__._quit() # pylint: disable=protected-access

    def _check_for_interrupt(self) -> bool:
        '''Checks for interrupts. Is called by GLib.timeout_add()'''
        if SIGNAL_HANDLER.interrupt_requested:
            LOGGER.info('Interrupt requested, shutting down...')
            self._cancel_pull()
            self.__class__._quit() # pylint: disable=protected-access
            return False # stop repeating
        return True # keep running

class SignalHandler:
    '''Thread-safe signal handler for interrupt requests (e.g., Ctrl+C).

    This class provides a safe way to check and set an interrupt flag
    from multiple threads.  It is designed to work with GLib’s signal
    handling (e.g., `GLib.unix_signal_add()`) while preventing race
    conditions during access.

    Attributes:
        interrupt_requested (bool):
            Read-only property indicating whether an interrupt
            (e.g., SIGINT/Ctrl+C) was triggered. Thread-safe.
    '''
    def __init__(self) -> None:
        '''Init the signal handler class.'''
        self._interrupt_requested: bool = False
        self._lock = threading.Lock()

    @property
    def interrupt_requested(self) -> bool:
        '''Check if an interrupt was requested with thread-safe
        read-only access.
        '''
        with self._lock:
            return self._interrupt_requested

    def handle_sigint(self, *_args: Any) -> bool:
        ''' Callback for Ctrl+C (SIGINT).

        :return: bool: True to keep the handler alive,
                       False to remove it.
        '''
        LOGGER.info('SIGINT')
        with self._lock:
            self._interrupt_requested = True
        return True

    def handle_sigterm(self, *_args: Any) -> bool:
        ''' Callback for Ctrl+C (SIGTERM).

        :return: bool: True to keep the handler alive,
                       False to remove it.
        '''
        LOGGER.info('SIGTERM')
        with self._lock:
            self._interrupt_requested = True
        return True

SIGNAL_HANDLER: SignalHandler = SignalHandler()

def quit_glib_main_loop(
        signum: int, _frame: Optional[FrameType] = None) -> None:
    '''Signal handler for signals from Python’s signal module

    :param signum: The signal number
    :param _frame:  Almost never used (it’s for debugging).
    '''
    if signum is not None:
        try:
            signal_name = signal.Signals(signum).name
        except ValueError: # In case signum isn't in Signals enum
            signal_name = str(signum)
        LOGGER.info('Received signal %s (%s), exiting...', signum, signal_name)
    if glib_main_loop is not None:
        glib_main_loop.quit()
    else:
        raise RuntimeError('glib_main_loop not initialized!')


if __name__ == '__main__':
    if not _ARGS.debug:
        LOG_HANDLER_NULL = logging.NullHandler()
    else:
        LOG_HANDLER_STREAM = logging.StreamHandler(stream=sys.stdout)
        LOG_FORMATTER = logging.Formatter(
            '%(asctime)s %(filename)s '
            'line %(lineno)d %(funcName)s %(levelname)s: '
            '%(message)s')
        LOG_HANDLER_STREAM.setFormatter(LOG_FORMATTER)
        LOGGER.setLevel(logging.DEBUG)
        LOGGER.addHandler(LOG_HANDLER_STREAM)

    try:
        locale.setlocale(locale.LC_ALL, '')
    except locale.Error:
        LOGGER.error("IBUS-WARNING **: Using the fallback 'C' locale")
        locale.setlocale(locale.LC_ALL, 'C')

    LOCALEDIR = os.getenv('IBUS_LOCALEDIR')
    gettext.bindtextdomain(DOMAINNAME, LOCALEDIR)

    if _ARGS.model is not None:
        OLLAMA_PULL_UI = OllamaPullUI(model=_ARGS.model)
        glib_main_loop = GLib.MainLoop()
        # signal.signal(signal.SIGTERM, quit_glib_main_loop) # kill <pid>
        # Ctrl+C (optional, can also use try/except KeyboardInterrupt)
        # signal.signal(signal.SIGINT, quit_glib_main_loop)
        GLib.unix_signal_add(
            GLib.PRIORITY_DEFAULT,
            signal.SIGTERM,
            SIGNAL_HANDLER.handle_sigterm, # kill <pid>
            None)
        GLib.unix_signal_add(
            GLib.PRIORITY_DEFAULT,
            signal.SIGINT,
            SIGNAL_HANDLER.handle_sigint, # Keyboard interrupt
            None)
        try:
            glib_main_loop.run()
        except KeyboardInterrupt:
            # SIGNINT (Control+C) received
            LOGGER.info('Control+C pressed, exiting ...')
            glib_main_loop.quit()



# vim:et sts=4 sw=4
#
# ibus-typing-booster - A completion input method for IBus
#
# Copyright (c) 2020 Mike FABIAN <mfabian@redhat.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>

'''
A test program to test input purpose and hints
'''

from typing import Dict
from typing import Any
from typing import Optional
from types import FrameType
import sys
import signal
import logging
import logging.handlers

from gi import require_version
require_version('GLib', '2.0')
# pylint: disable=wrong-import-position
from gi.repository import GLib # type: ignore

# set_prgname before importing other modules to show the name in warning
# messages when import modules are failed. E.g. Gtk.
GLib.set_application_name('InputPurposeTest')
# This makes gnome-shell load the .desktop file when running under Wayland:
GLib.set_prgname('InputPurposeTest')

# pylint: disable=import-error
sys.path = [sys.path[0]+'/../engine'] + sys.path
import itb_util_core
import itb_util_gui
from itb_gtk import Gtk, GTK_MAJOR
from g_compat_helpers import (
    add_child,
    show_all,
)
# pylint: enable=import-error
# pylint: enable=wrong-import-position

LOGGER = logging.getLogger('ibus-typing-booster')

glib_main_loop: Optional[GLib.MainLoop] = None

class InputPurposeTest(Gtk.Window):  # type: ignore[misc]
    '''
    User interface of the setup tool
    '''
    def __init__(self) -> None:
        Gtk.Window.__init__(self, title='Input Purpose Test')
        self.set_name('InputPurposeTest')
        self.set_modal(False)
        self.set_title('Input Purpose Test')
        if GTK_MAJOR >= 4:
            self.connect('close_request', self.on_close)
        else:
            self.connect('delete-event', self.on_close)

        main_container = Gtk.Box()
        main_container.set_orientation(Gtk.Orientation.VERTICAL)
        main_container.set_spacing(0)
        add_child(self, main_container)

        margin = 5

        self._input_purpose = itb_util_gui.InputPurpose.FREE_FORM
        input_purpose_combobox = Gtk.ComboBox()
        input_purpose_combobox.set_margin_start(margin)
        input_purpose_combobox.set_margin_end(margin)
        input_purpose_combobox.set_margin_top(margin)
        input_purpose_combobox.set_margin_bottom(margin)
        self._input_purpose_store = Gtk.ListStore(str, int)
        for purpose in list(itb_util_gui.InputPurpose):
            self._input_purpose_store.append([purpose.name, purpose])
        input_purpose_combobox.set_model(self._input_purpose_store)
        renderer_text = Gtk.CellRendererText()
        input_purpose_combobox.pack_start(renderer_text, True)
        input_purpose_combobox.add_attribute(renderer_text, "text", 0)
        for i, item in enumerate(self._input_purpose_store):
            if self._input_purpose == item[1]:
                input_purpose_combobox.set_active(i)
        input_purpose_combobox.connect(
            'changed', self.on_input_purpose_combobox_changed)

        add_child(main_container, input_purpose_combobox)

        self._input_hints = itb_util_gui.InputHints.NONE

        input_hints_checkbuttons: Dict[str, Gtk.CheckButton] = {}
        for hint in itb_util_gui.InputHints:
            if hint.name is None or hint.name == 'NONE':
                continue
            input_hints_checkbuttons[hint.name] = Gtk.CheckButton(
                label=hint.name)
            input_hints_checkbuttons[hint.name].set_margin_start(margin)
            input_hints_checkbuttons[hint.name].set_margin_end(margin)
            input_hints_checkbuttons[hint.name].set_margin_top(margin)
            input_hints_checkbuttons[hint.name].set_margin_bottom(margin)
            input_hints_checkbuttons[hint.name].set_active(False)
            input_hints_checkbuttons[hint.name].set_hexpand(False)
            input_hints_checkbuttons[hint.name].set_vexpand(False)
            input_hints_checkbuttons[hint.name].connect(
                'toggled', self.on_checkbutton, hint)
            add_child(main_container, input_hints_checkbuttons[hint.name])

        self._test_entry = Gtk.Entry()
        self._test_entry.set_margin_start(margin)
        self._test_entry.set_margin_end(margin)
        self._test_entry.set_margin_top(margin)
        self._test_entry.set_margin_bottom(margin)
        self._test_entry.set_visible(True)
        self._test_entry.set_can_focus(True)
        self._test_entry.set_hexpand(False)
        self._test_entry.set_vexpand(False)
        self._test_entry.set_input_purpose(self._input_purpose)
        self._test_entry.set_input_hints(self._input_hints)
        self._test_entry.connect('notify::text', self.on_test_entry)

        add_child(main_container, self._test_entry)

        self._test_text_view = Gtk.TextView()
        self._test_text_view.set_wrap_mode(Gtk.WrapMode.WORD_CHAR)
        margin = 10
        self._test_text_view.set_margin_start(margin)
        self._test_text_view.set_margin_end(margin)
        self._test_text_view.set_margin_top(margin)
        self._test_text_view.set_margin_bottom(margin)
        self._test_text_view_buffer = Gtk.TextBuffer()
        self._test_text_view.set_buffer(self._test_text_view_buffer)
        self._test_text_view.set_visible(True)
        self._test_text_view.set_can_focus(True)
        self._test_text_view.set_hexpand(False)
        self._test_text_view.set_vexpand(True)
        self._test_text_view.set_input_purpose(self._input_purpose)
        self._test_text_view.set_input_hints(self._input_hints)
        self._test_text_view_buffer.connect(
            'changed', self.on_test_text_view_buffer_changed)

        add_child(main_container, self._test_text_view)

        show_all(self)

    def on_close(self, *_args: Any) -> None: # pylint: disable=no-self-use
        '''Main window has been closed, quit the glib main loop'''
        LOGGER.info('Window deleted by the window manager.')
        if glib_main_loop is not None:
            glib_main_loop.quit()
        else:
            raise RuntimeError("glib_main_loop not initialized!")

    def on_test_entry( # pylint: disable=no-self-use
            self, widget: Gtk.Entry, _property_spec: Any) -> None:
        '''
        Called when something in the test entry has changed
        '''
        LOGGER.info('Test entry contains: “%s”', widget.get_text())

    def on_test_text_view_buffer_changed( # pylint: disable=no-self-use
            self, widget: Gtk.TextBuffer) -> None:
        '''
        Called when something in the test entry has changed
        '''
        LOGGER.info('Test text view contains: “%s”',
                    widget.get_text(
                        widget.get_start_iter(),
                        widget.get_end_iter(),
                        True))

    def on_input_purpose_combobox_changed(
            self, widget: Gtk.ComboBox) -> None:
        '''
        The combobox to choose the input purpose has been changed.
        '''
        tree_iter = widget.get_active_iter()
        if tree_iter is not None:
            model = widget.get_model()
            self._input_purpose = model[tree_iter][1]
            if self._input_purpose not in list(itb_util_gui.InputPurpose):
                LOGGER.info(
                    'self._input_purpose = %s (Unknown)',
                    self._input_purpose)
                return
            for input_purpose in list(itb_util_gui.InputPurpose):
                if self._input_purpose == input_purpose:
                    LOGGER.info(
                        'self._input_purpose = %s (%s)',
                        self._input_purpose, str(input_purpose))
                    self._test_entry.set_input_purpose(self._input_purpose)
                    self._test_text_view.set_input_purpose(self._input_purpose)
                    input_purpose_entry = (
                        self._test_entry.get_input_purpose())
                    input_purpose_text_view = (
                        self._test_text_view.get_input_purpose())
                    LOGGER.info(
                        'Input purpose changed to %s (%s)',
                        input_purpose_entry, str(input_purpose_entry))
                    if input_purpose_entry != input_purpose_text_view:
                        LOGGER.error(
                            'input_purpose_entry != '
                            'input_purpose_text_view: %s %s',
                            input_purpose_entry, input_purpose_text_view)
                    if input_purpose_entry != self._input_purpose:
                        LOGGER.error(
                            'input_purpose_entry != '
                            'self._input_purpose: %s %s',
                            input_purpose_entry, self._input_purpose)

    def on_checkbutton(
            self, widget: Gtk.CheckButton, hint: int) -> None:
        '''
        One of the check buttons to activate or deactivate an input hint
        has been clicked.
        '''
        LOGGER.info('Clicked checkbutton %s %s', widget, hint)
        if widget.get_active():
            self._input_hints |= hint
        else:
            self._input_hints &= ~hint
        self._test_entry.set_input_hints(Gtk.InputHints(self._input_hints))
        self._test_text_view.set_input_hints(Gtk.InputHints(self._input_hints))
        input_hints_entry = self._test_entry.get_input_hints()
        input_hints_text_view = self._test_text_view.get_input_hints()
        LOGGER.info('New value of self._input_hints=%s',
                    format(int(input_hints_entry), '016b'))
        if int(input_hints_entry) != int(self._input_hints):
            LOGGER.error(
                'input_hints_entry != self._input_hints: %s %s',
                input_hints_entry, self._input_hints)
        if int(input_hints_entry) != int(input_hints_text_view):
            LOGGER.error(
                'input_hints_entry != input_hints_text_view: %s %s',
                input_hints_entry, input_hints_text_view)
        for input_hint in list(itb_util_gui.InputHints):
            if self._input_hints & input_hint:
                LOGGER.info(
                    'hint: %s %s',
                    str(hint), format(int(hint), '016b'))

def quit_glib_main_loop(
        signum: int, _frame: Optional[FrameType] = None) -> None:
    '''Signal handler for signals from Python’s signal module

    :param signum: The signal number
    :param _frame:  Almost never used (it’s for debugging).
    '''
    if signum is not None:
        try:
            signal_name = signal.Signals(signum).name
        except ValueError: # In case signum isn't in Signals enum
            signal_name = str(signum)
        LOGGER.info('Received signal %s (%s), exiting...', signum, signal_name)
    if glib_main_loop is not None:
        glib_main_loop.quit()
    else:
        raise RuntimeError("glib_main_loop not initialized!")

if __name__ == '__main__':
    LOG_HANDLER_STREAM = logging.StreamHandler(stream=sys.stdout)
    LOG_FORMATTER = logging.Formatter(
        '%(asctime)s %(filename)s '
        'line %(lineno)d %(funcName)s %(levelname)s: '
        '%(message)s')
    LOG_HANDLER_STREAM.setFormatter(LOG_FORMATTER)
    LOGGER.setLevel(logging.DEBUG)
    LOGGER.addHandler(LOG_HANDLER_STREAM)
    LOGGER.info('********** STARTING **********')

    itb_util_core.set_program_name('inputpurposetes') # only 15 characters

    INPUT_PURPOSE_TEST = InputPurposeTest()
    glib_main_loop = GLib.MainLoop()
    signal.signal(signal.SIGTERM, quit_glib_main_loop) # kill <pid>
    # Ctrl+C (optional, can also use try/except KeyboardInterrupt)
    # signal.signal(signal.SIGINT, quit_glib_main_loop)
    try:
        glib_main_loop.run()
    except KeyboardInterrupt:
        # SIGNINT (Control+C) received
        LOGGER.info('Control+C pressed, exiting ...')
        glib_main_loop.quit()



# vim:et sts=4 sw=4
#
# ibus-typing-booster - A completion input method for IBus
#
# Copyright (c) 2011-2013 Anish Patil <apatil@redhat.com>
# Copyright (c) 2012-2018 Mike FABIAN <mfabian@redhat.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
'''
Main program of ibus-typing-booster
'''
from typing import Any
from typing import Union
from typing import Type
from typing import Optional
import types
import os
import sys
import argparse
import re
import logging
import logging.handlers
from signal import signal, SIGTERM, SIGINT
# pylint: disable=wrong-import-position
from gi import require_version
require_version('IBus', '1.0')
from gi.repository import IBus  # ty: ignore[unresolved-import]
require_version('Gio', '2.0')
from gi.repository import Gio # type: ignore
require_version('GLib', '2.0')
from gi.repository import GLib # type: ignore
# pylint: enable=wrong-import-position

import itb_util_core
import itb_version

LOGGER = logging.getLogger('ibus-typing-booster')

def log_unhandled_exception(
    exc_type: Type[BaseException],
    exc_value: BaseException,
    exc_traceback: Optional[types.TracebackType]
) -> None:
    '''Log any unhandled exception.'''
    if issubclass(exc_type, KeyboardInterrupt):
        sys.__excepthook__(exc_type, exc_value, exc_traceback)
        return
    LOGGER.error(
        'Unhandled exception',
        exc_info=(exc_type, exc_value, exc_traceback)
    )

sys.excepthook = log_unhandled_exception

# pylint: disable=invalid-name
DEBUG_LEVEL = 0
try:
    DEBUG_LEVEL = int(str(os.getenv('IBUS_TYPING_BOOSTER_DEBUG_LEVEL')))
except (TypeError, ValueError):
    DEBUG_LEVEL = 0
# pylint: enable=invalid-name

if os.getenv('IBUS_TYPING_BOOSTER_LOCATION'):
    ICON_DIR = os.path.join(
        str(os.getenv('IBUS_TYPING_BOOSTER_LOCATION')),
        'icons')
else:
    ICON_DIR = os.path.join(
        itb_version.get_prefix(), 'share/ibus-typing-booster/icons')

if os.getenv('IBUS_TYPING_BOOSTER_LIB_LOCATION'):
    SETUP_TOOL = os.path.join(
        str(os.getenv('IBUS_TYPING_BOOSTER_LIB_LOCATION')),
        'ibus-setup-typing-booster')
else:
    SETUP_TOOL = os.path.join(
        itb_version.get_prefix(),
        'libexec/ibus-setup-typing-booster')

def parse_args() -> Any:
    '''Parse the command line arguments'''
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--daemon', '-d',
        action='store_true',
        dest='daemon',
        default=False,
        help='Run as daemon, default: %(default)s')
    parser.add_argument(
        '--ibus', '-i',
        action='store_true',
        dest='ibus',
        default=False,
        help='Set the IME icon file, default: %(default)s')
    parser.add_argument(
        '--xml', '-x',
        action='store_true',
        dest='xml',
        default=False,
        help='output the engines xml part, default: %(default)s')
    parser.add_argument(
        '--no-debug', '-n',
        action='store_true',
        dest='no_debug',
        default=False,
        help='Do not write log file '
        + '~/.local/share/ibus-typing-booster/debug.log, '
        + 'default: %(default)s')
    parser.add_argument(
        '--profile', '-p',
        action='store_true',
        dest='profile',
        default=False,
        help='print profiling information into the debug log. '
        + 'Works only when --no-debug is not used.')
    return parser.parse_args()

_ARGS = parse_args()

if _ARGS.profile:
    import cProfile
    import pstats
    import io
    _PROFILE = cProfile.Profile()

if _ARGS.xml:
    from xml.etree.ElementTree import Element, SubElement, tostring
else:
    # Avoid importing factory when the --xml option is used because
    # factory imports other stuff which imports Gtk and that needs a
    # display.
    #
    # But by moving the import of factory here, it is possible to
    # use the --xml option in an environment where there is no
    # display without getting an error message like this:
    #
    #    $ env -u DISPLAY python3 main.py --xml
    #    Unable to init server: Could not connect: Connection refused
    #
    # The --xml option is used by “ibus write-cache” which is used
    # during rpm updates and then there is often no display and
    # the above error message appears.
    #
    # See: https://bugzilla.redhat.com/show_bug.cgi?id=1711646
    import factory

class IMApp:
    '''Input method application class'''
    def __init__(self, exec_by_ibus: bool) -> None:
        if DEBUG_LEVEL > 1:
            LOGGER.debug('IMApp.__init__(exec_by_ibus=%s)\n', exec_by_ibus)
        self.__mainloop = GLib.MainLoop()
        self.__bus = IBus.Bus()
        self.__bus.connect("disconnected", self.__bus_destroy_cb)
        # pylint: disable=possibly-used-before-assignment
        self.__factory = factory.EngineFactory(self.__bus)
        # pylint: enable=possibly-used-before-assignment
        self.destroyed = False
        if exec_by_ibus:
            self.__bus.request_name(
                "org.freedesktop.IBus.IbusTypingBooster", 0)
        else:
            self.__component = IBus.Component(
                name="org.freedesktop.IBus.IbusTypingBooster",
                description="Typing Booster Component",
                version=itb_version.get_version(),
                license="GPL",
                author=("Mike FABIAN <mfabian@redhat.com>, "
                        + "Anish Patil <anish.developer@gmail.com>"),
                homepage="http://mike-fabian.github.io/ibus-typing-booster",
                textdomain="ibus-typing-booster")
            name = 'typing-booster'
            longname = 'Typing Booster'
            description = 'A completion input method to speedup typing.'
            language = 't'
            author = (
                'Mike FABIAN <mfabian@redhat.com>'
                + ', Anish Patil <anish.developer@gmail.com>')
            icon = os.path.join(ICON_DIR, 'ibus-typing-booster.svg')
            if not os.access(icon, os.F_OK):
                icon = ''
            layout = 'default'
            symbol = '🚀'

            engine = IBus.EngineDesc(name=name,
                                     longname=longname,
                                     description=description,
                                     language=language,
                                     license='GPL',
                                     author=author,
                                     icon=icon,
                                     layout=layout,
                                     symbol=symbol)
            self.__component.add_engine(engine)
            self.__bus.register_component(self.__component)

    def run(self) -> None:
        '''Run the input method application'''
        if DEBUG_LEVEL > 1:
            LOGGER.debug('IMApp.run()\n')
        if _ARGS.profile:
            _PROFILE.enable()
        self.__mainloop.run()
        self.__bus_destroy_cb()

    def quit(self) -> None:
        '''Quit the input method application'''
        if DEBUG_LEVEL > 1:
            LOGGER.debug('IMApp.quit()\n')
        self.__bus_destroy_cb()

    def __bus_destroy_cb(self, bus: Any = None) -> None:
        if DEBUG_LEVEL > 1:
            LOGGER.debug('IMApp.__bus_destroy_cb(bus=%s)\n', bus)
        if self.destroyed:
            return
        LOGGER.info('finalizing:)')
        self.__factory.do_destroy()
        self.destroyed = True
        self.__mainloop.quit()
        if _ARGS.profile:
            _PROFILE.disable()
            stats_stream = io.StringIO()
            stats = pstats.Stats(_PROFILE, stream=stats_stream)
            stats.strip_dirs()
            stats.sort_stats('cumulative')
            stats.print_stats('tabsqlite', 25)
            stats.print_stats('hunspell_suggest', 25)
            stats.print_stats('hunspell_table', 25)
            stats.print_stats('itb_util_core', 25)
            stats.print_stats('itb_util_gui', 25)
            stats.print_stats('itb_emoji', 25)
            LOGGER.info('Profiling info:\n%s', stats_stream.getvalue())

def cleanup(ima_ins: IMApp) -> None:
    '''
    Clean up when the input method application was killed by a signal
    '''
    ima_ins.quit()
    sys.exit()

def indent(element: Any, level: int = 0) -> None:
    '''Use to format xml Element pretty :)'''
    i = "\n" + level*"    "
    if element is None:
        return
    if len(element):
        if not element.text or not element.text.strip():
            element.text = i + "    "
        last_subelement = None
        for subelement in element:
            last_subelement = subelement
            indent(subelement, level+1)
            if not subelement.tail or not subelement.tail.strip():
                subelement.tail = i + "    "
        if (last_subelement is not None
            and (not last_subelement.tail or not last_subelement.tail.strip())):
            last_subelement.tail = i
    else:
        if level and (not element.tail or not element.tail.strip()):
            element.tail = i

def write_xml() -> None:
    '''
    Writes the XML to describe the engine(s) to standard output.
    '''
    m17n_db_info = itb_util_core.M17nDbInfo()
    m17n_input_methods = set(m17n_db_info.get_imes())
    m17n_input_methods_skip = set([
        'NoIME',
    ])
    supported_input_methods = sorted(
        m17n_input_methods - m17n_input_methods_skip
    ) + ['typing-booster']

    indic_languages = (
        'as',
        'bn',
        'brx',
        'doi',
        'dra',
        'dv',
        'gu',
        'hi',
        'kn',
        'ks',
        'mai',
        'ml',
        'mni',
        'mr',
        'ne',
        'new',
        'or',
        'pa',
        'sa',
        'sat',
        'sd',
        'si',
        'ta',
        'te',
    )
    rank_patterns = [ # first matching regexp wins
        # Use higher ranks than ibus-m17n does for the same
        # input methods to have the “tb” input methods appear
        # with higher priority than the “m17n” input methods.
        #
        # Assign lower rank to Sinhala Samanala since it is a
	# non-keyboard input method in Sinhala.  (Not sure whether
	# this makes sense but ibus-m17n does it).
        (r'tb:si:samanala', 0),
        # Assign higher rank to all inscript2 engines since they
        # should be preferred over the inscript engines.
        # ibus-m17n uses rank 3 for these.
        (r'tb:[^:]+:.*inscript2.*', 4),
        # Assign higher rank to Indic engines which represent each
	# language. ibus-m17n uses rank 2 for these.
        (r'tb:te:inscript', 3),
        (r'tb:ta:tamil99', 3),
        (r'tb:si:wijesekara', 3),
        (r'tb:sd:inscript', 3),
        (r'tb:sa:harvard-kyoto', 3),
        (r'tb:pa:inscript', 3),
        (r'tb:or:inscript', 3),
        (r'tb:ne:rom', 3),
        (r'tb:mr:inscript', 3),
        (r'tb:ml:inscript', 3),
        (r'tb:mai:inscript', 3),
        (r'tb:ks:kbd', 3),
        (r'tb:kn:kgp', 3),
        (r'tb:hi:inscript', 3),
        (r'tb:gu:inscript', 3),
        (r'tb:bn:inscript', 3),
        (r'tb:as:inscript', 3),
        # Indic engines should be selected by default:
	# https://bugzilla.redhat.com/show_bug.cgi?id=640896
        # ibus-m17n uses rank 1 for these.
        (r'tb:(' + ('|').join(indic_languages) + r'):[^:]+', 2),
        # Arabic kbd engine should be selected by default:
        # https://bugzilla.redhat.com/show_bug.cgi?id=1076945
        # ibus-m17n uses rank 1 for the ar-kbd engine.
        (r'tb:ar:kbd', 2),
        (r'tb:ja:anthy', -1), # ja:anthy is superseded by ibus-anthy
        (r'tb:zh:py', -1), # zh:py is superseded by ibus-pinyin etc
        (r'tb:zh:py-gb', -1), # zh:py-gb is superseded by ibus-pinyin etc
        (r'tb:zh:py-b5', -1), # zh:py-b5 is superseded by ibus-pinyin etc
        (r'tb:ko:han2', -1), # ko:han2 is superseded by ibus-hangul
        (r'tb:ko:romaja', -1), # ko:romaja is superseded by ibus-hangul
    ]
    # Some Indic engines expect AltGr is automatically mapped
    ralt_switch_patterns = [
        r'tb:[^:]+:.*inscript.*',
        r'tb:si:.+',
    ]

    egs = Element('engines') # pylint: disable=possibly-used-before-assignment

    recent_gnome = (itb_util_core.is_desktop('gnome')
                    and itb_util_core.get_gnome_shell_version() >= (48, 3))

    for ime in supported_input_methods:
        _engine = SubElement( # pylint: disable=possibly-used-before-assignment
            egs, 'engine')

        name = 'typing-booster'
        longname = 'Typing Booster'
        language = 't'
        engine_license = 'GPL'
        author = ('Mike FABIAN <mfabian@redhat.com>'
                  ', Anish Patil <anish.developer@gmail.com>')
        layout = 'default'
        # For something like “setxkbmap 'us(intl)'”, layout is “us”
        # and layout_variant is “intl”.
        layout_variant = ''
        layout_option = ''
        icon = os.path.join(ICON_DIR, 'ibus-typing-booster.svg')
        description = 'A completion input method to speedup typing.'
        symbol = '🚀'
        if recent_gnome:
            # If running on Gnome and gnome-shell is new enough to contain
            # https://gitlab.gnome.org/GNOME/gnome-shell/-/merge_requests/3753
            # make the symbol black and white:
            symbol = '🚀\uFE0E'
        icon_prop_key = 'InputMode'
        rank = 0
        schema_path = '/org/freedesktop/ibus/engine/typing-booster/'
        if ime != 'typing-booster':
            m17n_lang, m17n_name = ime.split('-', maxsplit=1)
            name = f'tb:{m17n_lang}:{m17n_name}'
            schema_path = ('/org/freedesktop/ibus/engine/tb/'
                           f'{m17n_lang}/{m17n_name}/')
            rank = 0
            for pattern, pattern_rank in rank_patterns:
                if re.fullmatch(pattern, name):
                    rank = pattern_rank
                    break # first matching regexp wins
            layout = 'default'
            for pattern in ralt_switch_patterns:
                if re.fullmatch(pattern, name):
                    # I do this only to match the xml output of
                    # ibus-m17n, it does not seem to help though,
                    # i.e. it does not automatically make the right
                    # alt key work as ISO_Level3_Shift when switching
                    # from a keyboard layout like “English (US)” which
                    # does not haave this:
                    layout_option = 'lv3:ralt_switch'
                    break
            symbol = ''
            for pattern, pattern_symbol in itb_util_core.M17N_IME_SYMBOLS:
                if re.fullmatch(pattern, name):
                    symbol = pattern_symbol
                    break
            longname = f'{ime} (tb)'
            language = m17n_lang
            icon =  m17n_db_info.get_icon(ime)
            description = m17n_db_info.get_description(ime)
        setup = SETUP_TOOL + f' --engine-name {name}'
        gsettings = Gio.Settings(
            schema='org.freedesktop.ibus.engine.typing-booster',
            path=schema_path)
        user_symbol = itb_util_core.variant_to_value(
            gsettings.get_user_value('inputmodetruesymbol'))
        if user_symbol is not None:
            symbol = user_symbol

        _name = SubElement(_engine, 'name')
        _name.text = name
        _longname = SubElement(_engine, 'longname')
        _longname.text = longname
        _language = SubElement(_engine, 'language')
        _language.text = language
        _license = SubElement(_engine, 'license')
        _license.text = engine_license
        _author = SubElement(_engine, 'author')
        _author.text = author
        _icon = SubElement(_engine, 'icon')
        _icon.text = icon
        _layout = SubElement(_engine, 'layout')
        _layout.text = layout
        _layout_variant = SubElement(_engine, 'layout_variant')
        _layout_variant.text = layout_variant
        _layout_option = SubElement(_engine, 'layout_option')
        _layout_option.text = layout_option
        _desc = SubElement(_engine, 'description')
        _desc.text = description
        _symbol = SubElement(_engine, 'symbol')
        _symbol.text = symbol
        _setup = SubElement(_engine, 'setup')
        _setup.text = setup
        _icon_prop_key = SubElement(_engine, 'icon_prop_key')
        _icon_prop_key.text = icon_prop_key
        _rank = SubElement(_engine, 'rank')
        _rank.text = str(rank)

    # now format the xmlout pretty
    indent(egs)
    egsout = tostring( # pylint: disable=possibly-used-before-assignment
        egs, encoding='utf8', method='xml').decode('utf-8')
    patt = re.compile(r'<\?.*\?>\n')
    egsout = patt.sub('', egsout)
    sys.stdout.buffer.write((egsout+'\n').encode('utf-8'))

def main() -> None:
    '''Main program'''
    if _ARGS.xml:
        write_xml()
        return

    log_handler: Union[
        logging.NullHandler, logging.handlers.TimedRotatingFileHandler] = (
            logging.NullHandler())
    if not _ARGS.no_debug:
        directory = os.path.expanduser('~/.local/share/ibus-typing-booster')
        logfile = os.path.join(directory, 'debug.log')
        try:
            os.makedirs(directory, exist_ok=True)
            log_handler = logging.handlers.TimedRotatingFileHandler(
                logfile,
                when='midnight',
                interval=1,
                backupCount=7,
                encoding='UTF-8',
                delay=False,
                utc=False,
                atTime=None)
        except OSError as error:
            sys.stderr.write(
                f'Warning: cannot initialize debug log file {logfile}: {error}\n')
            log_handler = logging.NullHandler()
    log_formatter = logging.Formatter(
        '%(asctime)s %(filename)s '
        'line %(lineno)d %(funcName)s %(levelname)s: '
        '%(message)s')
    log_handler.setFormatter(log_formatter)
    LOGGER.setLevel(logging.DEBUG)
    LOGGER.addHandler(log_handler)
    LOGGER.info('********** STARTING **********')
    itb_util_core.set_program_name('ibus-engine-tb')
    if _ARGS.daemon:
        if os.fork():
            sys.exit()

    ima = IMApp(_ARGS.ibus)
    signal(SIGTERM, lambda signum, stack_frame: cleanup(ima))
    signal(SIGINT, lambda signum, stack_frame: cleanup(ima))
    try:
        ima.run()
    except KeyboardInterrupt:
        ima.quit()
    return

if __name__ == "__main__":
    main()



# vim:et sts=4 sw=4
#
# ibus-typing-booster - A completion input method for IBus
#
# Copyright (c) 2012-2013 Anish Patil <apatil@redhat.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
'''
Module to initialize gettext
'''
import locale
import gettext
import os

DOMAINNAME = 'ibus-typing-booster'

def _(text: str) -> str:
    '''Gettext translation function.'''
    return gettext.dgettext(DOMAINNAME, text)

def N_(text: str) -> str: # pylint: disable=invalid-name
    '''Mark string for translation without actually translating.

    Used by gettext tools to extract strings that need translation.
    '''
    return text

def init() -> None:
    '''
    Initialize gettext
    '''
    localedir = os.getenv("IBUS_LOCALEDIR")
    # Python's locale module doesn't provide all methods on some
    # operating systems like FreeBSD
    try:
        # for non-standard localedir
        locale.bindtextdomain(DOMAINNAME, localedir)
    except AttributeError:
        pass
    gettext.bindtextdomain(DOMAINNAME, localedir)
